# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
This moduel implements a TokenInterface for `CoAP over WebSockets`_.

.. _`CoAP over WebSockets`: https://tools.ietf.org/html/rfc8323#section-4

As with CoAP-over-TCP, while the transport distinguishes a connection initiator
("WebSocket (and TCP) client") and a receiver ("WebSocket (and TCP) server"),
both sides can take both roles in CoAP (ie. as a CoAP server and a CoAP
client). As the WebSocket client can not possibly be connected to (even by the
same server -- once the connection is closed, it's gone and even a new one
likely has a different port), aiocoap does not allow expressing their addresses
in URIs (given they wouldn't serve their purpose as URLs and don't provide any
stability either). Requests to a CoAP-over-WS client can be made by assigning
the remote to an outgoing request.

Port choice
-----------

Unlike the other transports, CoAP-over-WS is specified with a privileged port
(port 80) as the default port. This is impractical for aiocoap servers for two
reasons:

    * Unless explicitly configured, aiocoap is typically run as an unprivileged
      user (and has no provisions in place to receive a socket by other means
      than opening it).

    * Where a CoAP-over-WS proxy is run, there is often a "proper" website
      running on the same port on a full HTTP server. That server is usually
      capable of forwarding requests, whereas the ``websockets`` module used by
      aiocoap is in no position to either serve websites nor to proxy to an
      underlying server.

The recommended setup is therefore to run a full web server at port 80, and
configure it to proxy incoming requests for WebSockets at `/.well-known/coap`
to aiocoap's server, which defaults to binding to port 8683.

The port choice of outgoing connections, or the interpretation of the
protocol's default port (ie. the port implied by ``coap+ws://hostname/``) is of
course unaffected by this.

.. warning::

  Due to a shortcoming of aiocoap's way of specifying ports to bind
  to, if a port is explicitly stated to bind to, CoAP-over-WS will bind to that
  port plus 3000 (resulting in the abovementioned 8683 for 5683). If TLS server
  keys are given, the TLS server is launched on the next port after the HTTP
  server (typically 8684).

Using on pyodide_
-----------------

When use  on pyodide_,
instead of using the ``websockets`` module,
a simplified client-only back-end is used,
which utilizes the browser's WebSocket API.

.. _pyodide: https://pyodide.org/
"""

from __future__ import annotations

from typing import Dict, List
from collections import namedtuple
import asyncio
import functools
import http
import weakref

from aiocoap import Message, interfaces, ABORT, util, error
from aiocoap.transports import rfc8323common
from ..util.asyncio import py38args
from ..defaults import is_pyodide

if is_pyodide:
    import aiocoap.util.pyodide_websockets as websockets
else:
    import websockets

def _decode_message(data: bytes) -> Message:
    codeoffset = 1
    tokenoffset = 2

    tkl = data[0]
    if tkl > 8:
        raise error.UnparsableMessage("Overly long token")
    code = data[codeoffset]
    token = data[tokenoffset:tokenoffset + tkl]

    msg = Message(code=code, token=token)

    msg.payload = msg.opt.decode(data[tokenoffset + tkl:])

    return msg

def _serialize(msg: Message) -> bytes:
    tkl = len(msg.token)
    if tkl > 8:
        raise ValueError("Overly long token")

    data = [
            bytes((tkl, msg.code,)),
            msg.token,
            msg.opt.encode(),
            ]
    if msg.payload:
        data += [b'\xff', msg.payload]

    return b"".join(data)

PoolKey = namedtuple("PoolKey", ("scheme", "hostinfo"))

class WSRemote(rfc8323common.RFC8323Remote, interfaces.EndpointAddress):
    _connection: websockets.WebSocketCommonProtocol
    # Only used to ensure that remotes are associated to the right pool -- not
    # that there'd be any good reason to have multiple of those.
    _pool: weakref.ReferenceType[WSPool]

    scheme = None # Override property -- it's per instance here

    def __init__(self, pool, connection, loop, log, *, scheme, local_hostinfo=None, remote_hostinfo=None):
        super().__init__()
        self._pool = weakref.ref(pool)
        self._connection = connection
        self.loop = loop
        self.log = log

        self._local_is_server = isinstance(connection, websockets.WebSocketServerProtocol)

        if local_hostinfo is None:
            self._local_hostinfo = self._connection.local_address[:2]
        else:
            self._local_hostinfo = local_hostinfo
        if remote_hostinfo is None:
            self._remote_hostinfo = self._connection.remote_address[:2]
        else:
            self._remote_hostinfo = remote_hostinfo

        self.scheme = scheme

        # Goes both for client and for server ends; on the server end, it
        # ensures that role reversal URIs can be used even when passed as URIs
        # and not as remotes (although that's of course only possible locally).
        self._poolkey = PoolKey(self.scheme, self.hostinfo)

    # Necessary for RFC8323Remote

    def _abort_with(self, msg, *, close_code=1002):
        # Like _send_message, this may take actual time -- but unlike there,
        # there's no need to regulate back-pressure
        self.loop.create_task(
                self._abort_with_waiting(msg, close_code=close_code),
                **py38args(name="Abortion WebSocket sonnection with %r" % msg)
                )

    # Unlike _send_message, this is pulled out of the the _abort_with function
    # as it's also used in _run_recv_loop
    async def _abort_with_waiting(self, msg, *, close_code):
        self.log.debug("Aborting with message: %r", msg)
        try:
            await self._connection.send(_serialize(msg))
        except Exception as e:
            self.log.error("Sending to a WebSocket should not raise errors", exc_info=e)
        await self._connection.close(code=close_code)

    def _send_message(self, msg):
        # FIXME overhaul back-pressure model
        async def send():
            self.log.debug("Sending message: %r", msg)
            try:
                await self._connection.send(_serialize(msg))
            except Exception as e:
                self.log.error("Sending to a WebSocket should not raise errors", exc_info=e)
        self.loop.create_task(
                send(),
                **py38args(name="WebSocket sending of %r" % msg)
                )

    async def release(self):
        await super().release()
        try:
            await self._connection.wait_closed()
        except asyncio.CancelledError:
            self.log.warning(
                    "Connection %s was not closed by peer in time after release",
                    self
                    )

class WSPool(interfaces.TokenInterface):
    _outgoing_starting: Dict[PoolKey, asyncio.Task]
    _pool: Dict[PoolKey, WSRemote]

    _servers: List[websockets.WebSocketServer]

    def __init__(self, tman, log, loop):
        self.loop = loop

        self._pool = {}
        self._outgoing_starting = {}

        self._servers = []

        # See where it is used for documentation, remove when not needed any more
        self._in_shutdown = False

        self._tokenmanager = tman
        self.log = log

    @classmethod
    async def create_transport(cls, tman: interfaces.TokenManager, log, loop, *, client_credentials, server_bind=None, server_context=None):
        self = cls(tman, log, loop)

        self._client_credentials = client_credentials

        if server_bind:
            host, port = server_bind
            if port is None:
                port = 8683
            elif port != 0:
                # FIXME see module documentation
                port = port + 3000

            server = await websockets.serve(
                    functools.partial(self._new_connection, scheme='coap+ws'),
                    host, port,
                    subprotocols=['coap'],
                    process_request=self._process_request,
                    ping_interval=None, # "SHOULD NOT be used"
                    )
            self._servers.append(server)

            if server_context is not None:
                server = await websockets.serve(
                        functools.partial(self._new_connection, scheme='coaps+ws'),
                        host, port + 1,
                        subprotocols=['coap'],
                        process_request=self._process_request,
                        ping_interval=None, # "SHOULD NOT be used"
                        ssl=server_context,
                        )
                self._servers.append(server)

        return self

    # Helpers for WebScoket server

    async def _new_connection(self, websocket, path=None, *, scheme):
        # ignoring path: Already checked in _process_request
        #
        # (path is present up to 10.0 and absent in 10.1; keeping it around to
        # stay compatible with different versions).

        hostheader = websocket.request_headers['Host']
        if hostheader.count(':') > 1 and '[' not in hostheader:
            # Workaround for websockets version before
            # https://github.com/aaugustin/websockets/issues/802
            #
            # To be removed once a websockets version with this fix can be
            # depended on
            hostheader = '[' + hostheader[:hostheader.rfind(':')] + ']' + hostheader[hostheader.rfind(':'):]
        local_hostinfo = util.hostportsplit(hostheader)

        remote = WSRemote(self, websocket, self.loop, self.log, scheme=scheme, local_hostinfo=local_hostinfo)
        self._pool[remote._poolkey] = remote

        await self._run_recv_loop(remote)

    @staticmethod
    async def _process_request(path, request_headers):
        if path != '/.well-known/coap':
            return (http.HTTPStatus.NOT_FOUND, [], b"")
        # Continue with WebSockets
        return None

    # Helpers for WebScoket client

    def _connect(self, key: PoolKey):
        self._outgoing_starting[key] = self.loop.create_task(
                self._connect_task(key),
                **py38args(name="WebSocket connection opening to %r" % (key,))
                )

    async def _connect_task(self, key: PoolKey):
        try:
            ssl_context = self._client_credentials.ssl_client_context(key.scheme, key.hostinfo)

            hostinfo_split = util.hostportsplit(key.hostinfo)

            ws_scheme = {'coap+ws': 'ws', 'coaps+ws': 'wss'}[key.scheme]

            if ws_scheme == 'ws' and ssl_context is not None:
                raise ValueError("An SSL context was provided for a remote accessed via a plaintext websockets.")
            if ws_scheme == 'wss':
                if is_pyodide:
                    if ssl_context is not None:
                        raise ValueError("The pyodide websocket implementation can't be configured with a non-default context -- connections created in a browser will always use the browser's CA set.")
                else:
                    if ssl_context is None:
                        # Like with TLSClient, we need to pass in a default
                        # context and can't rely on the websocket library to
                        # use a default one when wss is requested (it doesn't)
                        import ssl
                        ssl_context = ssl.create_default_context()

            websocket = await websockets.connect("%s://%s/.well-known/coap" % (
                ws_scheme, key.hostinfo),
                subprotocols=['coap'],
                ping_interval=None,
                ssl=ssl_context,
                )

            remote = WSRemote(self, websocket, self.loop, self.log, scheme=key.scheme, remote_hostinfo=hostinfo_split)
            assert remote._poolkey == key, "Pool key construction is inconsistent"
            self._pool[key] = remote

            self.loop.create_task(
                    self._run_recv_loop(remote),
                    **py38args(name="WebSocket receive loop for %r" % (key,))
                    )

            return remote
        finally:
            del self._outgoing_starting[key]

    # Implementation of TokenInterface

    async def fill_or_recognize_remote(self, message):
        if isinstance(message.remote, WSRemote) and \
                message.remote._pool() is self:
            return True

        if message.requested_scheme in ('coap+ws', 'coaps+ws'):
            key = PoolKey(message.requested_scheme, message.remote.hostinfo)

            if key in self._pool:
                message.remote = self._pool[key]
                if message.remote._connection.open:
                    return True
                # else try opening a new one

            if key not in self._outgoing_starting:
                self._connect(key)
            # It's a bit unorthodox to wait for an (at least partially)
            # established connection in fill_or_recognize_remote, but it's
            # not completely off off either, and it makes it way easier to
            # not have partially initialized remotes around
            message.remote = await self._outgoing_starting[key]
            return True

        return False

    def send_message(self, message, messageerror_monitor):
        # Ignoring messageerror_monitor: CoAP over reliable transports has no
        # way of indicating that a particular message was bad, it always shuts
        # down the complete connection

        if message.code.is_response():
            no_response = (message.opt.no_response or 0) & (1 << message.code.class_ - 1) != 0
            if no_response:
                return

        message.opt.no_response = None

        message.remote._send_message(message)

    async def shutdown(self):
        self._in_shutdown = True
        self.log.debug("Shutting down any connections on %r", self)

        client_shutdowns = [
            asyncio.create_task(
                c.release(),
                **py38args(name="Close connection %s" % c)
            )
            for c in self._pool.values()]

        server_shutdowns = []
        while self._servers:
            s = self._servers.pop()
            # We could do something like
            # >>> for websocket in s.websockets:
            # >>>     del websocket.logger.extra['websocket']
            # to reduce the reference loops
            # (websocket.logger.extra['websocket'] == websocket), but as the
            # tests actually do run a GC collection once and that gets broken
            # up, it's not worth adding fragilty here
            s.close()
            server_shutdowns.append(asyncio.create_task(
                s.wait_closed(),
                **py38args(name="Close server %s" % s)))

        # Placing client shutdowns before server shutdowns to give them a
        # chance to send out Abort messages; the .close() method could be more
        # helpful here by stopping new connections but letting us finish off
        # the old ones
        shutdowns = client_shutdowns + server_shutdowns
        if shutdowns:
            # wait is documented to require a non-empty set
            await asyncio.wait(shutdowns)

    # Incoming message processing

    async def _run_recv_loop(self, remote):
        remote._send_initial_csm()

        while True:
            try:
                received = await remote._connection.recv()
            except websockets.exceptions.ConnectionClosed:
                # This check is purely needed to silence the warning printed
                # from tokenmanager, "Internal shutdown sequence msismatch:
                # error dispatched through tokenmanager after shutdown" -- and
                # is a symptom of https://github.com/chrysn/aiocoap/issues/284
                # and of the odd circumstance that we can't easily cancel the
                # _run_recv_loop tasks (as we should while that issue is
                # unresolved) in the shutdown handler.
                if not self._in_shutdown:
                    # FIXME if deposited somewhere, mark that as stale?
                    self._tokenmanager.dispatch_error(error.RemoteServerShutdown("Peer closed connection"), remote)
                return

            if not isinstance(received, bytes):
                await remote._abort_with_waiting(Message(code=ABORT, payload=b"Text frame received"), close_code=1003)
                return

            try:
                msg = _decode_message(received)
            except error.UnparsableMessage:
                await remote._abort_with_waiting(Message(code=ABORT, payload=b"Message parsing error"), close_code=1007)
                return

            msg.remote = remote

            if msg.code.is_signalling():
                try:
                    remote._process_signaling(msg)
                except rfc8323common.CloseConnection as e:
                    self._tokenmanager.dispatch_error(e.args[0], msg.remote)
                    self._pool.pop(remote._poolkey)
                    await remote._connection.close()
                continue

            if remote._remote_settings is None:
                remote.abort("No CSM received")
                return

            if msg.code.is_response():
                self._tokenmanager.process_response(msg)
                # ignoring the return value; unexpected responses can be the
                # asynchronous result of cancelled observations
            else:
                self._tokenmanager.process_request(msg)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
CoAP-over-TLS transport (early work in progress)

Right now this is running on self-signed, hard-coded certificates with default
SSL module options.

To use this, generate keys as with::

    $ openssl req -x509 -newkey rsa:4096 -keyout key.pem -out cert.pem -days 5 -nodes

and state your hostname (eg. localhost) when asked for the Common Name.
"""

from .tcp import TCPClient, TCPServer

from aiocoap import COAPS_PORT

class _TLSMixIn:
    _scheme = 'coaps+tcp'
    _default_port = COAPS_PORT

class TLSServer(_TLSMixIn, TCPServer):
    @classmethod
    async def create_server(cls, bind, tman, log, loop, server_context):
        return await super().create_server(bind, tman, log, loop, _server_context=server_context)

class TLSClient(_TLSMixIn, TCPClient):
    def _ssl_context_factory(self, hostinfo):
        c = self.credentials.ssl_client_context(self._scheme, hostinfo)
        if c is None:
            import ssl
            c = ssl.create_default_context()
        c.set_alpn_protocols(["coap"])
        return c



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module implements a MessageInterface for UDP based on the asyncio
DatagramProtocol.

This is a simple version that works only for clients (by creating a dedicated
unbound but connected socket for each communication partner) and probably not
with multicast (it is assumed to be unsafe for multicast), which can be
expected to work even on platforms where the :mod:`.udp6` module can not be
made to work (Android, OSX, Windows for missing ``recvmsg`` and socket options,
or any event loops that don't have an add_reader method).

Note that the name of the module is a misnomer (and the module is likely to be
renamed): Nothing in it is IPv6 specific; the socket is created using whichever
address family the OS chooses based on the given host name.

One small but noteworthy detail about this transport is that it does not
distinguish between IP literals and host names. As a result, requests and
responses from remotes will appear to arrive from a remote whose netloc is the
requested name, not an IP literal.

This transport is experimental, likely to change, and not fully tested yet
(because the test suite is not yet ready to matrix-test the same tests with
different transport implementations, and because it still fails in proxy
blockwise tests).

For one particular use case, this may be usable for servers in a sense: If (and
only if) all incoming requests are only ever sent from clients that were
previously addressed as servers by the running instance. (This is generally
undesirable as it greatly limits the usefulness of the server, but is used in
LwM2M setups). As such a setup makes demands on the peer that are not justified
by the CoAP specification (in particular, that it send requests from a
particular port), this should still only be used for cases where the udp6
transport is unavailable due to platform limitations.
"""

import asyncio
from collections import OrderedDict
import socket

from aiocoap import error
from aiocoap import interfaces
from aiocoap import COAP_PORT
from ..util import hostportjoin
from ..util.asyncio import py38args
from .generic_udp import GenericMessageInterface

class _Connection(asyncio.DatagramProtocol, interfaces.EndpointAddress):
    def __init__(self, ready_callback, message_interface: "GenericMessageInterface", stored_sockaddr):
        self._ready_callback = ready_callback
        self._message_interface = message_interface

        # This gets stored in the _Connection because not all implementations
        # of datagram transports will expose the get_extra_info('socket')
        # (right now, all I knew do), or their backend might not be a connected
        # socket (like in uvloop), so the information can't be just obtained
        # from the transport, but is needed to implement .hostinfo
        #
        # If _Connections become used in other contexts (eg. tinydtls starts
        # using them), it might be a good idea to move all this into a subclass
        # and split it from the pure networking stuff.
        self.hostinfo = hostportjoin(stored_sockaddr[0], None if stored_sockaddr[1] == COAP_PORT else stored_sockaddr[1])

        self._stage = "initializing" #: Status property purely for debugging

    def __repr__(self):
        return "<%s at %#x on transport %s, %s>" % (
                type(self).__name__,
                id(self),
                getattr(self, "_transport", "(none)"),
                self._stage)

    # address interface

    is_multicast = False

    is_multicast_locally = False

    scheme = 'coap'

    # statically initialized in init
    hostinfo = None
    uri_base = None
    uri_base = property(lambda self: 'coap://' + self.hostinfo)

    @property
    def hostinfo_local(self):
        # This can only be done on a best-effort base here. Unlike the below
        # hostinfo (see comments there), there is no easy way around this, so
        # if there are still implementations out that don't do the extras,
        # that's it and the calling site should reconsider whether they need
        # something that can not be determined. (Some more effort could go into
        # falling back to get_extra_info('socket').getsockname(), but that
        # should really be fixed in the transport provider).
        if not hasattr(self, '_transport'):
            raise RuntimeError("Simple6 does not have defined local host info in current stage %s" % self._stage)
        sockname = self._transport.get_extra_info('sockname')
        if sockname is None:
            raise RuntimeError("Simple6 can not determine local address from the underlying UDP implementation")
        return hostportjoin(*sockname[:2])
    uri_base_local = property(lambda self: 'coap://' + self.hostinfo_local)

    @property
    def blockwise_key(self):
        # Not pulling in hostinfo_local as that's unreliable anyway -- if we
        # change identity and the (UDP sever) follows its requests across our
        # changing port, that's probably fine.
        return self.hostinfo

# fully disabled because some implementations of asyncio don't make the
# information available; going the easy route and storing it for all (see
# attribute population in __init__)

#     # FIXME continued: probably this is, and the above is not (or should not be)
#     @property
#     def hostinfo(self):
#         print("ACCESSING HOSTINFO")
#         host, port = self._transport.get_extra_info('socket').getpeername()[:2]
#         if port == COAP_PORT:
#             port = None
#         # FIXME this should use some of the _plainaddress mechanisms of the udp6 addresses
#         return hostportjoin(host, port)

    # datagram protocol interface

    def connection_made(self, transport):
        self._transport = transport
        self._ready_callback()
        self._stage = "active"
        del self._ready_callback

    def datagram_received(self, data, address):
        self._message_interface._received_datagram(self, data)

    def error_received(self, exception):
        self._message_interface._received_exception(self, exception)

    def connection_lost(self, exception):
        if exception is None:
            pass
        else:
            self._new_error_callback(self, exception)

    # whatever it is _DatagramClientSocketpoolSimple6 expects

    # ... because generic_udp expects it from _DatagramClientSocketpoolSimple6
    def send(self, data):
        self._transport.sendto(data, None)

    async def shutdown(self):
        self._stage = "shutting down"
        self._transport.abort()
        del self._message_interface
        self._stage = "destroyed"

class _DatagramClientSocketpoolSimple6:
    """This class is used to explore what an Python/asyncio abstraction around
    a hypothetical "UDP connections" mechanism could look like.

    Assume there were a socket variety that had UDP messages (ie. unreliable,
    unordered, boundary-preserving) but that can do an accept() like a TCP
    listening socket can, and can create outgoing connection-ish sockets from
    the listeing port.

    That interface would be usable for all UDP-based CoAP transport
    implementations; this particular implementation, due to limitations of
    POSIX sockets (and the additional limitations imposed on it like not using
    PKTINFO) provides the interface, but only implements the outgoing part, and
    will not allow setting the outgoing port or interface."""

    max_sockets = 64

    # FIXME (new_message_callback, new_error_callback) should probably rather
    # be one object with a defined interface; either that's the
    # MessageInterfaceSimple6 and stored accessibly (so the Protocol can know
    # which MessageInterface to talk to for sending), or we move the
    # MessageInterface out completely and have that object be the Protocol,
    # and the Protocol can even send new packages via the address
    def __init__(self, loop, mi: "GenericMessageInterface"):
        # using an OrderedDict to implement an LRU cache as it's suitable for that purpose according to its documentation
        self._sockets = OrderedDict()

        self._loop = loop
        self._message_interface = mi

    async def _maybe_purge_sockets(self):
        while len(self._sockets) >= self.max_sockets: # more of an if
            oldaddr, oldest = next(iter(self._sockets.items()))
            await oldest.shutdown()
            del self._sockets[oldaddr]

    async def connect(self, sockaddr):
        """Create a new socket with a given remote socket address

        Note that the sockaddr does not need to be fully resolved or complete,
        as it is not used for matching incoming packages; ('host.example.com',
        5683) is perfectly OK (and will create a different outgoing socket that
        ('hostalias.example.com', 5683) even if that has the same address, for
        better or for worse).

        For where the general underlying interface is concerned, it is not yet
        fixed at all when this must return identical objects."""

        protocol = self._sockets.get(sockaddr)
        if protocol is not None:
            self._sockets.move_to_end(sockaddr)
            return protocol

        await self._maybe_purge_sockets()

        ready = asyncio.get_running_loop().create_future()
        try:
            transport, protocol = await self._loop.create_datagram_endpoint(
                    lambda: _Connection(lambda: ready.set_result(None), self._message_interface, sockaddr),
                    remote_addr=sockaddr)
        except socket.gaierror as e:
            raise error.ResolutionError("No address information found for requests to %r" % (sockaddr,)) from e
        await ready

#         # Enable this to easily make every connection to localhost a new one
#         # during testing
#         import random
#         sockaddr = sockaddr + (random.random(),)

        # FIXME twice: 1., those never get removed yet (should timeout or
        # remove themselves on error), and 2., this is racy against a shutdown right after a connect
        self._sockets[sockaddr] = protocol

        return protocol

    async def shutdown(self):
        # preventing the creation of new sockets early on, and generally
        # breaking cycles
        del self._message_interface

        if self._sockets:
            done, pending = await asyncio.wait([
                asyncio.create_task(
                    s.shutdown(),
                    **py38args(name="Socket shutdown of %r" % s)
                    )
                for s
                in self._sockets.values()])
            for item in done:
                await item
        del self._sockets

class MessageInterfaceSimple6(GenericMessageInterface):
    @classmethod
    async def create_client_transport_endpoint(cls, ctx, log, loop):
        self = cls(ctx, log, loop)

        # Cyclic reference broken during shutdown
        self._pool = _DatagramClientSocketpoolSimple6(self._loop, self)
        return self

    async def recognize_remote(self, remote):
        return isinstance(remote, _Connection) and remote in self._pool._sockets



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Constants either defined in the CoAP protocol (often default values for lack
of ways to determine eg. the estimated round trip time). Some parameters are
invented here for practical purposes of the implementation (eg.
DEFAULT_BLOCK_SIZE_EXP, EMPTY_ACK_DELAY)."""

import warnings
import string

COAP_PORT = 5683
"""The IANA-assigned standard port for COAP services."""

COAPS_PORT = 5684

MCAST_IPV4_ALLCOAPNODES = "224.0.1.187"
MCAST_IPV6_LINKLOCAL_ALLNODES = "ff02::1"
MCAST_IPV6_LINKLOCAL_ALLCOAPNODES = "ff02::fd"
MCAST_IPV6_SITELOCAL_ALLNODES = "ff05::1"
MCAST_IPV6_SITELOCAL_ALLCOAPNODES = "ff05::fd"
MCAST_ALL = (
      MCAST_IPV4_ALLCOAPNODES,
      MCAST_IPV6_LINKLOCAL_ALLNODES,
      MCAST_IPV6_LINKLOCAL_ALLCOAPNODES,
      MCAST_IPV6_SITELOCAL_ALLNODES,
      MCAST_IPV6_SITELOCAL_ALLCOAPNODES,
      )

MAX_REGULAR_BLOCK_SIZE_EXP = 6

class TransportTuning:
    """Base parameters that guide CoAP transport behaviors

    The values in here are recommended values, often defaults from RFCs. They
    can be tuned in subclasses (and then passed into a message as
    ``transport_tuning``), although users should be aware that alteing some of
    these can cause the library to behave in ways violating the specification,
    especially with respect to congestion control.
    """

    #   +-------------------+---------------+
    #   | name              | default value |
    #   +-------------------+---------------+
    #   | ACK_TIMEOUT       | 2 seconds     |
    #   | ACK_RANDOM_FACTOR | 1.5           |
    #   | MAX_RETRANSMIT    | 4             |
    #   | NSTART            | 1             |
    #   | DEFAULT_LEISURE   | 5 seconds     |
    #   | PROBING_RATE      | 1 Byte/second |
    #   +-------------------+---------------+

    ACK_TIMEOUT = 2.0
    """The time, in seconds, to wait for an acknowledgement of a
    confirmable message. The inter-transmission time doubles
    for each retransmission."""

    ACK_RANDOM_FACTOR = 1.5
    """Timeout multiplier for anti-synchronization."""

    MAX_RETRANSMIT = 4
    """The number of retransmissions of confirmable messages to
    non-multicast endpoints before the infrastructure assumes no
    acknowledgement will be received."""

    NSTART = 1
    """Maximum number of simultaneous outstanding interactions
       that endpoint maintains to a given server (including proxies)"""

    #   +-------------------+---------------+
    #   | name              | default value |
    #   +-------------------+---------------+
    #   | MAX_TRANSMIT_SPAN |          45 s |
    #   | MAX_TRANSMIT_WAIT |          93 s |
    #   | MAX_LATENCY       |         100 s |
    #   | PROCESSING_DELAY  |           2 s |
    #   | MAX_RTT           |         202 s |
    #   | EXCHANGE_LIFETIME |         247 s |
    #   | NON_LIFETIME      |         145 s |
    #   +-------------------+---------------+

    @property
    def MAX_TRANSMIT_SPAN(self):
        """Maximum time from the first transmission
        of a confirmable message to its last retransmission."""
        return self.ACK_TIMEOUT * (2 ** self.MAX_RETRANSMIT - 1) * self.ACK_RANDOM_FACTOR

    @property
    def MAX_TRANSMIT_WAIT(self):
        """Maximum time from the first transmission
        of a confirmable message to the time when the sender gives up on
        receiving an acknowledgement or reset."""
        return self.ACK_TIMEOUT * (2 ** (self.MAX_RETRANSMIT + 1) - 1) * self.ACK_RANDOM_FACTOR

    MAX_LATENCY = 100.0
    """Maximum time a datagram is expected to take from the start
    of its transmission to the completion of its reception."""

    @property
    def PROCESSING_DELAY(self):
        """"Time a node takes to turn around a
        confirmable message into an acknowledgement."""
        return self.ACK_TIMEOUT

    @property
    def MAX_RTT(self):
        """Maximum round-trip time."""
        return 2 * self.MAX_LATENCY + self.PROCESSING_DELAY

    @property
    def EXCHANGE_LIFETIME(self):
        """time from starting to send a confirmable message to the time when an
        acknowledgement is no longer expected, i.e. message layer information about the
        message exchange can be purged"""
        return self.MAX_TRANSMIT_SPAN + self.MAX_RTT

    DEFAULT_BLOCK_SIZE_EXP = MAX_REGULAR_BLOCK_SIZE_EXP
    """Default size exponent for blockwise transfers."""

    EMPTY_ACK_DELAY = 0.1
    """After this time protocol sends empty ACK, and separate response"""

    REQUEST_TIMEOUT = MAX_TRANSMIT_WAIT
    """Time after which server assumes it won't receive any answer.
       It is not defined by IETF documents.
       For human-operated devices it might be preferable to set some small value
       (for example 10 seconds)
       For M2M it's application dependent."""

    DEFAULT_LEISURE = 5

    @property
    def MULTICAST_REQUEST_TIMEOUT(self):
        return self.REQUEST_TIMEOUT + self.DEFAULT_LEISURE

    OBSERVATION_RESET_TIME = 128
    """Time in seconds after which the value of the observe field are ignored.

    This number is not explicitly named in RFC7641.
    """

_default_transport_tuning = TransportTuning()
def __getattr__(name):
    if name[0] in string.ascii_uppercase and hasattr(_default_transport_tuning, name):
        warnings.warn(f"{name} is deprecated, use through the message's transport_tuning instead", DeprecationWarning, stacklevel=2)
        return getattr(_default_transport_tuning, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")

SHUTDOWN_TIMEOUT = 3
"""Maximum time, in seconds, for which the process is kept around during shutdown"""

__all__ = [k for k in dir() if not k.startswith('_') and k not in ('warnings', 'strings')]



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
Common errors for the aiocoap library
"""

import warnings
import abc

from .numbers import codes
from . import util

class Error(Exception):
    """
    Base exception for all exceptions that indicate a failed request
    """

class RenderableError(Error, metaclass=abc.ABCMeta):
    """
    Exception that can meaningfully be represented in a CoAP response
    """

    @abc.abstractmethod
    def to_message(self):
        """Create a CoAP message that should be sent when this exception is
        rendered"""

class ResponseWrappingError(Error):
    """
    An exception that is raised due to an unsuccessful but received response.

    A better relationship with :mod:`.numbers.codes` should be worked out to do
    ``except UnsupportedMediaType`` (similar to the various ``OSError``
    subclasses).
    """
    def __init__(self, coapmessage):
        self.coapmessage = coapmessage

    def to_message(self):
        return self.coapmessage

    def __repr__(self):
        return "<%s: %s %r>" % (type(self).__name__, self.coapmessage.code, self.coapmessage.payload)

class ConstructionRenderableError(RenderableError):
    """
    RenderableError that is constructed from class attributes :attr:`code` and
    :attr:`message` (where the can be overridden in the constructor).
    """

    def __init__(self, message=None):
        if message is not None:
            self.message = message

    def to_message(self):
        from .message import Message
        return Message(code=self.code, payload=self.message.encode('utf8'))

    code = codes.INTERNAL_SERVER_ERROR #: Code assigned to messages built from it
    message = "" #: Text sent in the built message's payload

# FIXME: this should be comprehensive, maybe generted from the code list

class NotFound(ConstructionRenderableError):
    code = codes.NOT_FOUND

class MethodNotAllowed(ConstructionRenderableError):
    code = codes.METHOD_NOT_ALLOWED

class UnsupportedContentFormat(ConstructionRenderableError):
    code = codes.UNSUPPORTED_CONTENT_FORMAT

class Unauthorized(ConstructionRenderableError):
    code = codes.UNAUTHORIZED

class BadRequest(ConstructionRenderableError):
    code = codes.BAD_REQUEST

# More detailed versions of code based errors

class NoResource(NotFound):
    """
    Raised when resource is not found.
    """
    message = "Error: Resource not found!"
    def __init__(self):
        warnings.warn("NoResource is deprecated in favor of NotFound", DeprecationWarning, stacklevel=2)

class UnallowedMethod(MethodNotAllowed):
    """
    Raised by a resource when request method is understood by the server
    but not allowed for that particular resource.
    """
    message = "Error: Method not allowed!"

class UnsupportedMethod(MethodNotAllowed):
    """
    Raised when request method is not understood by the server at all.
    """
    message = "Error: Method not recognized!"


class NetworkError(Error):
    """Base class for all "something went wrong with name resolution, sending
    or receiving packages".

    Errors of these kinds are raised towards client callers when things went
    wrong network-side, or at context creation. They are often raised from
    socket.gaierror or similar classes, but these are wrapped in order to make
    catching them possible independently of the underlying transport."""

class ResolutionError(NetworkError):
    """Resolving the host component of a URI to a usable transport address was
    not possible"""

class MessageError(NetworkError):
    """Received an error from the remote on the CoAP message level (typically a
    RST)"""

class NotImplemented(Error):
    """
    Raised when request is correct, but feature is not implemented
    by library.
    For example non-sequential blockwise transfers
    """

class RemoteServerShutdown(NetworkError):
    """The peer a request was sent to in a stateful connection closed the
    connection around the time the request was sent"""

class TimeoutError(NetworkError):
    """Base for all timeout-ish errors.

    Like NetworkError, receiving this alone does not indicate whether the
    request may have reached the server or not.
    """

class ConRetransmitsExceeded(TimeoutError):
    """A transport that retransmits CON messages has failed to obtain a response
    within its retransmission timeout.

    When this is raised in a transport, requests failing with it may or may
    have been received by the server.
    """

class RequestTimedOut(TimeoutError):
    """
    Raised when request is timed out.

    This error is currently not produced by aiocoap; it is deprecated. Users
    can now catch error.TimeoutError, or newer more detailed subtypes
    introduced later.
    """


class WaitingForClientTimedOut(TimeoutError):
    """
    Raised when server expects some client action:

        - sending next PUT/POST request with block1 or block2 option
        - sending next GET request with block2 option

    but client does nothing.

    This error is currently not produced by aiocoap; it is deprecated. Users
    can now catch error.TimeoutError, or newer more detailed subtypes
    introduced later.
    """

class ResourceChanged(Error):
    """
    The requested resource was modified during the request and could therefore
    not be received in a consistent state.
    """

class UnexpectedBlock1Option(Error):
    """
    Raised when a server responds with block1 options that just don't match.
    """

class UnexpectedBlock2(Error):
    """
    Raised when a server responds with another block2 than expected.
    """

class MissingBlock2Option(Error):
    """
    Raised when response with Block2 option is expected
    (previous response had Block2 option with More flag set),
    but response without Block2 option is received.
    """

class NotObservable(Error):
    """
    The server did not accept the request to observe the resource.
    """

class ObservationCancelled(Error):
    """
    The server claimed that it will no longer sustain the observation.
    """

class UnparsableMessage(Error):
    """
    An incoming message does not look like CoAP.

    Note that this happens rarely -- the requirements are just two bit at the
    beginning of the message, and a minimum length.
    """

class LibraryShutdown(Error):
    """The library or a transport registered with it was requested to shut
    down; this error is raised in all outstanding requests."""

class AnonymousHost(Error):
    """This is raised when it is attempted to express as a reference a (base)
    URI of a host or a resource that can not be reached by any process other
    than this.

    Typically, this happens when trying to serialize a link to a resource that
    is hosted on a CoAP-over-TCP or -WebSockets client: Such resources can be
    accessed for as long as the connection is active, but can not be used any
    more once it is closed or even by another system."""

__getattr__ = util.deprecation_getattr({
        "UnsupportedMediaType": "UnsupportedContentFormat",
        "RequestTimedOut": "TimeoutError",
        "WaitingForClientTimedOut": "TimeoutError",
        }, globals())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import urllib.parse
import struct
import copy
import string
from collections import namedtuple

from . import error, optiontypes
from .numbers.codes import Code, CHANGED
from .numbers.types import Type
from .numbers.constants import TransportTuning
from .options import Options
from .util import hostportjoin, hostportsplit, Sentinel, quote_nonascii
from .util.uri import quote_factory, unreserved, sub_delims
from . import interfaces

__all__ = ['Message', 'NoResponse']

# FIXME there should be a proper inteface for this that does all the urllib
# patching possibly required and works with pluggable transports. urls qualify
# if they can be parsed into the Proxy-Scheme / Uri-* structure.
coap_schemes = ['coap', 'coaps', 'coap+tcp', 'coaps+tcp', 'coap+ws', 'coaps+ws']

# Monkey patch urllib to make URL joining available in CoAP
# This is a workaround for <http://bugs.python.org/issue23759>.
urllib.parse.uses_relative.extend(coap_schemes)
urllib.parse.uses_netloc.extend(coap_schemes)

class Message(object):
    """CoAP Message with some handling metadata

    This object's attributes provide access to the fields in a CoAP message and
    can be directly manipulated.

    * Some attributes are additional data that do not round-trip through
      serialization and deserialization. They are marked as "non-roundtrippable".
    * Some attributes that need to be filled for submission of the message can
      be left empty by most applications, and will be taken care of by the
      library. Those are marked as "managed".

    The attributes are:

    * :attr:`payload`: The payload (body) of the message as bytes.
    * :attr:`mtype`: Message type (CON, ACK etc, see :mod:`.numbers.types`).
      Managed unless set by the application.
    * :attr:`code`: The code (either request or response code), see
      :mod:`.numbers.codes`.
    * :attr:`opt`: A container for the options, see :class:`.options.Options`.

    * :attr:`mid`: The message ID. Managed by the :class:`.Context`.
    * :attr:`token`: The message's token as bytes. Managed by the :class:`.Context`.
    * :attr:`remote`: The socket address of the other side, managed by the
      :class:`.protocol.Request` by resolving the ``.opt.uri_host`` or
      ``unresolved_remote``, or the :class:`.Responder` by echoing the incoming
      request's. Follows the :class:`.interfaces.EndpointAddress` interface.
      Non-roundtrippable.

      While a message has not been transmitted, the property is managed by the
      :class:`.Message` itself using the :meth:`.set_request_uri()` or the
      constructor `uri` argument.
    * :attr:`transport_tuning`: Parameters used by one or more transports to
      guide transmission. These are purely advisory hints; unknown properties
      of that object are ignored, and transports consider them over built-in
      constants on a best-effort basis.

      Note that many attributes are mandatory if this is not None; it is
      recommended that any objects passed in here are based on the
      :class:`aiocoap.numbers.constants.TransportTuning` class.

    * :attr:`request`: The request to which an incoming response message
      belongs; only available at the client. Managed by the
      :class:`.interfaces.RequestProvider` (typically a :class:`.Context`).

    These properties are still available but deprecated:

    * requested_*: Managed by the :class:`.protocol.Request` a response results
      from, and filled with the request's URL data. Non-roundtrippable.

    * unresolved_remote: ``host[:port]`` (strictly speaking; hostinfo as in a
      URI) formatted string. If this attribute is set, it overrides
      ``.RequestManageropt.uri_host`` (and ``-_port``) when it comes to filling the
      ``remote`` in an outgoing request.

      Use this when you want to send a request with a host name that would not
      normally resolve to the destination address. (Typically, this is used for
      proxying.)

    Options can be given as further keyword arguments at message construction
    time. This feature is experimental, as future message parameters could
    collide with options.


    The four messages involved in an exchange
    -----------------------------------------

    ::

        Requester                                  Responder

        +-------------+                          +-------------+
        | request msg |  ---- send request --->  | request msg |
        +-------------+                          +-------------+
                                                       |
                                                  processed into
                                                       |
                                                       v
        +-------------+                          +-------------+
        | response m. |  <--- send response ---  | response m. |
        +-------------+                          +-------------+


    The above shows the four message instances involved in communication
    between an aiocoap client and server process. Boxes represent instances of
    Message, and the messages on the same line represent a single CoAP as
    passed around on the network. Still, they differ in some aspects:

        * The requested URI will look different between requester and responder
          if the requester uses a host name and does not send it in the message.
        * If the request was sent via multicast, the response's requested URI
          differs from the request URI because it has the responder's address
          filled in. That address is not known at the responder's side yet, as
          it is typically filled out by the network stack.
        * It is yet unclear whether the response's URI should contain an IP
          literal or a host name in the unicast case if the Uri-Host option was
          not sent.
        * Properties like Message ID and token will differ if a proxy was
          involved.
        * Some options or even the payload may differ if a proxy was involved.
    """

    def __init__(self, *, mtype=None, mid=None, code=None, payload=b'', token=b'', uri=None, transport_tuning=None, **kwargs):
        self.version = 1
        if mtype is None:
            # leave it unspecified for convenience, sending functions will know what to do
            self.mtype = None
        else:
            self.mtype = Type(mtype)
        self.mid = mid
        if code is None:
            # as above with mtype
            self.code = None
        else:
            self.code = Code(code)
        self.token = token
        self.payload = payload
        self.opt = Options()

        self.remote = None

        self.transport_tuning = transport_tuning or TransportTuning()

        # deprecation error, should go away roughly after 0.2 release
        if self.payload is None:
            raise TypeError("Payload must not be None. Use empty string instead.")

        if uri:
            self.set_request_uri(uri)

        for k, v in kwargs.items():
            setattr(self.opt, k, v)

    def __repr__(self):
        return "<aiocoap.Message at %#x: %s %s (%s, %s) remote %s%s%s>" % (
                id(self),
                self.mtype if self.mtype is not None else "no mtype,",
                self.code,
                "MID %s" % self.mid if self.mid is not None else "no MID",
                "token %s" % self.token.hex() if self.token else "empty token",
                self.remote,
                ", %s option(s)" % len(self.opt._options) if self.opt._options else "",
                ", %s byte(s) payload" % len(self.payload) if self.payload else ""
                )

    def _repr_html_(self):
        import html
        if not self.payload:
            payload_rendered = '<p>No payload</p>'
        else:
            from . import defaults
            if defaults.prettyprint_missing_modules():
                payload_rendered = f"<code>{html.escape(repr(self.payload))}</code>"
            else:
                from .util.prettyprint import pretty_print, lexer_for_mime
                (notes, mediatype, text) = pretty_print(self)
                import pygments
                from pygments.formatters import HtmlFormatter
                try:
                    lexer = lexer_for_mime(mediatype)
                    text = pygments.highlight(text, lexer, HtmlFormatter())
                except pygments.util.ClassNotFound:
                    text = html.escape(text)
                payload_rendered = (
                        "<div>"
                        + "".join(f'<p style="color:gray;font-size:small;">{html.escape(n)}</p>' for n in notes)
                        + f"<pre>{text}</pre>"
                        + "</div>"
                        )
        return f"""<details style="padding-left:1em"><summary style="margin-left:-1em;display:list-item;">Message with code {self.code._repr_html_() if self.code is not None else 'None'}, remote {html.escape(str(self.remote))}</summary>
                {self.opt._repr_html_()}{payload_rendered}"""

    def copy(self, **kwargs):
        """Create a copy of the Message. kwargs are treated like the named
        arguments in the constructor, and update the copy."""
        # This is part of moving messages in an "immutable" direction; not
        # necessarily hard immutable. Let's see where this goes.

        new = type(self)(
                mtype=kwargs.pop('mtype', self.mtype),
                mid=kwargs.pop('mid', self.mid),
                code=kwargs.pop('code', self.code),
                payload=kwargs.pop('payload', self.payload),
                token=kwargs.pop('token', self.token),
                # Assuming these are not readily mutated, but rather passed
                # around in a class-like fashion
                transport_tuning=kwargs.pop('transport_tuning', self.transport_tuning),
                )
        new.remote = kwargs.pop('remote', self.remote)
        new.opt = copy.deepcopy(self.opt)

        if 'uri' in kwargs:
            new.set_request_uri(kwargs.pop('uri'))

        for k, v in kwargs.items():
            setattr(new.opt, k, v)

        return new

    @classmethod
    def decode(cls, rawdata, remote=None):
        """Create Message object from binary representation of message."""
        try:
            (vttkl, code, mid) = struct.unpack('!BBH', rawdata[:4])
        except struct.error:
            raise error.UnparsableMessage("Incoming message too short for CoAP")
        version = (vttkl & 0xC0) >> 6
        if version != 1:
            raise error.UnparsableMessage("Fatal Error: Protocol Version must be 1")
        mtype = (vttkl & 0x30) >> 4
        token_length = (vttkl & 0x0F)
        msg = Message(mtype=mtype, mid=mid, code=code)
        msg.token = rawdata[4:4 + token_length]
        msg.payload = msg.opt.decode(rawdata[4 + token_length:])
        msg.remote = remote
        return msg

    def encode(self):
        """Create binary representation of message from Message object."""
        if self.code is None or self.mtype is None or self.mid is None:
            raise TypeError("Fatal Error: Code, Message Type and Message ID must not be None.")
        rawdata = bytes([(self.version << 6) + ((self.mtype & 0x03) << 4) + (len(self.token) & 0x0F)])
        rawdata += struct.pack('!BH', self.code, self.mid)
        rawdata += self.token
        rawdata += self.opt.encode()
        if len(self.payload) > 0:
            rawdata += bytes([0xFF])
            rawdata += self.payload
        return rawdata

    def get_cache_key(self, ignore_options=()):
        """Generate a hashable and comparable object (currently a tuple) from
        the message's code and all option values that are part of the cache key
        and not in the optional list of ignore_options (which is the list of
        option numbers that are not technically NoCacheKey but handled by the
        application using this method).

        >>> from aiocoap.numbers import GET
        >>> m1 = Message(code=GET)
        >>> m2 = Message(code=GET)
        >>> m1.opt.uri_path = ('s', '1')
        >>> m2.opt.uri_path = ('s', '1')
        >>> m1.opt.size1 = 10 # the only no-cache-key option in the base spec
        >>> m2.opt.size1 = 20
        >>> m1.get_cache_key() == m2.get_cache_key()
        True
        >>> m2.opt.etag = b'000'
        >>> m1.get_cache_key() == m2.get_cache_key()
        False
        >>> from aiocoap.numbers.optionnumbers import OptionNumber
        >>> ignore = [OptionNumber.ETAG]
        >>> m1.get_cache_key(ignore) == m2.get_cache_key(ignore)
        True
        """

        options = []

        for option in self.opt.option_list():
            if option.number in ignore_options or (option.number.is_safetoforward() and option.number.is_nocachekey()):
                continue
            options.append((option.number, option.value))

        return (self.code, tuple(options))

    #
    # splitting and merging messages into and from message blocks
    #

    def _extract_block(self, number, size_exp, max_bert_size):
        """Extract block from current message."""
        if size_exp == 7:
            start = number * 1024
            size = 1024 * (max_bert_size // 1024)
        else:
            size = 2 ** (size_exp + 4)
            start = number * size

        if start >= len(self.payload):
            raise error.BadRequest("Block request out of bounds")

        end = start + size if start + size < len(self.payload) else len(self.payload)
        more = True if end < len(self.payload) else False

        payload = self.payload[start:end]
        blockopt = (number, more, size_exp)

        if self.code.is_request():
            return self.copy(
                    payload=payload,
                    mid=None,
                    block1=blockopt
                    )
        else:
            return self.copy(
                    payload=payload,
                    mid=None,
                    block2=blockopt
                    )

    def _append_request_block(self, next_block):
        """Modify message by appending another block"""
        if not self.code.is_request():
            raise ValueError("_append_request_block only works on requests.")

        block1 = next_block.opt.block1
        if block1.more:
            if len(next_block.payload) == block1.size:
                pass
            elif block1.size_exponent == 7 and \
                    len(next_block.payload) % block1.size == 0:
                pass
            else:
                raise error.BadRequest("Payload size does not match Block1")
        if block1.start == len(self.payload):
            self.payload += next_block.payload
            self.opt.block1 = block1
            self.token = next_block.token
            self.mid = next_block.mid
            if not block1.more and next_block.opt.block2 is not None:
                self.opt.block2 = next_block.opt.block2
        else:
            # possible extension point: allow messages with "gaps"; then
            # ValueError would only be raised when trying to overwrite an
            # existing part; it is doubtful though that the blockwise
            # specification even condones such behavior.
            raise ValueError()

    def _append_response_block(self, next_block):
        """Append next block to current response message.
           Used when assembling incoming blockwise responses."""
        if not self.code.is_response():
            raise ValueError("_append_response_block only works on responses.")

        block2 = next_block.opt.block2
        if not block2.is_valid_for_payload_size(len(next_block.payload)):
            raise error.UnexpectedBlock2("Payload size does not match Block2")
        if block2.start != len(self.payload):
            # Does not need to be implemented as long as the requesting code
            # sequentially clocks out data
            raise error.NotImplemented()

        if next_block.opt.etag != self.opt.etag:
            raise error.ResourceChanged()

        self.payload += next_block.payload
        self.opt.block2 = block2
        self.token = next_block.token
        self.mid = next_block.mid

    def _generate_next_block2_request(self, response):
        """Generate a sub-request for next response block.

        This method is used by client after receiving blockwise response from
        server with "more" flag set."""

        # Note: response here is the assembled response, but (due to
        # _append_response_block's workings) it carries the Block2 option of
        # the last received block.

        next_after_received = len(response.payload) // response.opt.block2.size
        blockopt = optiontypes.BlockOption.BlockwiseTuple(
                next_after_received, False, response.opt.block2.size_exponent)

        # has been checked in assembly, just making sure
        assert blockopt.start == len(response.payload), "Unexpected state of preassembled message"

        blockopt = blockopt.reduced_to(response.remote.maximum_block_size_exp)

        return self.copy(
                payload=b"",
                mid=None,
                token=None,
                block2=blockopt,
                block1=None,
                observe=None
                )

    def _generate_next_block1_response(self):
        """Generate a response to acknowledge incoming request block.

        This method is used by server after receiving blockwise request from
        client with "more" flag set."""
        response = Message(code=CHANGED, token=self.token)
        response.remote = self.remote
        if self.opt.block1.block_number == 0 and self.opt.block1.size_exponent > self.transport_tuning.DEFAULT_BLOCK_SIZE_EXP:
            new_size_exponent = self.transport_tuning.DEFAULT_BLOCK_SIZE_EXP
            response.opt.block1 = (0, True, new_size_exponent)
        else:
            response.opt.block1 = (self.opt.block1.block_number, True, self.opt.block1.size_exponent)
        return response

    #
    # the message in the context of network and addresses
    #


    def get_request_uri(self, *, local_is_server=False):
        """The absolute URI this message belongs to.

        For requests, this is composed from the options (falling back to the
        remote). For responses, this is largely taken from the original request
        message (so far, that could have been trackecd by the requesting
        application as well), but -- in case of a multicast request -- with the
        host replaced by the responder's endpoint details.

        This implements Section 6.5 of RFC7252.

        By default, these values are only valid on the client. To determine a
        message's request URI on the server, set the local_is_server argument
        to True. Note that determining the request URI on the server is brittle
        when behind a reverse proxy, may not be possible on all platforms, and
        can only be applied to a request message in a renderer (for the
        response message created by the renderer will only be populated when it
        gets transmitted; simple manual copying of the request's remote to the
        response will not magically make this work, for in the very case where
        the request and response's URIs differ, that would not catch the
        difference and still report the multicast address, while the actual
        sending address will only be populated by the operating system later).
        """

        # maybe this function does not belong exactly *here*, but it belongs to
        # the results of .request(message), which is currently a message itself.

        if hasattr(self, '_original_request_uri'):
            # During server-side processing, a message's options may be altered
            # to the point where its options don't accurately reflect its URI
            # any more. In that case, this is stored.
            return self._original_request_uri

        if self.code.is_response():
            refmsg = self.request

            if refmsg.remote.is_multicast:
                if local_is_server:
                    multicast_netloc_override = self.remote.hostinfo_local
                else:
                    multicast_netloc_override = self.remote.hostinfo
            else:
                multicast_netloc_override = None
        else:
            refmsg = self
            multicast_netloc_override = None

        proxyuri = refmsg.opt.proxy_uri
        if proxyuri is not None:
            return proxyuri

        scheme = refmsg.opt.proxy_scheme or refmsg.remote.scheme
        query = refmsg.opt.uri_query or ()
        path = refmsg.opt.uri_path

        if multicast_netloc_override is not None:
            netloc = multicast_netloc_override
        else:
            if local_is_server:
                netloc = refmsg.remote.hostinfo_local
            else:
                netloc = refmsg.remote.hostinfo

            if refmsg.opt.uri_host is not None or \
                    refmsg.opt.uri_port is not None:

                host, port = hostportsplit(netloc)

                host = refmsg.opt.uri_host or host
                port = refmsg.opt.uri_port or port

                # FIXME: This sounds like it should be part of
                # hpostportjoin/-split
                escaped_host = quote_nonascii(host)

                # FIXME: "If host is not valid reg-name / IP-literal / IPv4address,
                # fail"

                netloc = hostportjoin(escaped_host, port)

        # FIXME this should follow coap section 6.5 more closely
        query = "&".join(_quote_for_query(q) for q in query)
        path = ''.join("/" + _quote_for_path(p) for p in path) or '/'

        fragment = None
        params = "" # are they not there at all?

        # Eases debugging, for when they raise from urunparse you won't know
        # which of them it was
        assert scheme is not None, "Remote has no scheme set"
        assert netloc is not None, "Remote has no netloc set"
        return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))

    def set_request_uri(self, uri, *, set_uri_host=True):
        """Parse a given URI into the uri_* fields of the options.

        The remote does not get set automatically; instead, the remote data is
        stored in the uri_host and uri_port options. That is because name resolution
        is coupled with network specifics the protocol will know better by the
        time the message is sent. Whatever sends the message, be it the
        protocol itself, a proxy wrapper or an alternative transport, will know
        how to handle the information correctly.

        When ``set_uri_host=False`` is passed, the host/port is stored in the
        ``unresolved_remote`` message property instead of the uri_host option;
        as a result, the unresolved host name is not sent on the wire, which
        breaks virtual hosts but makes message sizes smaller.

        This implements Section 6.4 of RFC7252.
        """

        parsed = urllib.parse.urlparse(uri)

        if parsed.fragment:
            raise ValueError("Fragment identifiers can not be set on a request URI")

        if parsed.scheme not in coap_schemes:
            self.opt.proxy_uri = uri
            return

        if parsed.username or parsed.password:
            raise ValueError("User name and password not supported.")

        if parsed.path not in ('', '/'):
            self.opt.uri_path = [urllib.parse.unquote(x) for x in parsed.path.split('/')[1:]]
        else:
            self.opt.uri_path = []
        if parsed.query:
            self.opt.uri_query = [urllib.parse.unquote(x) for x in parsed.query.split('&')]
        else:
            self.opt.uri_query = []

        self.remote = UndecidedRemote(parsed.scheme, parsed.netloc)

        is_ip_literal = parsed.netloc.startswith('[') or (
                parsed.hostname.count('.') == 3 and
                all(c in '0123456789.' for c in parsed.hostname) and
                all(int(x) <= 255 for x in parsed.hostname.split('.')))

        if set_uri_host and not is_ip_literal:
            self.opt.uri_host = urllib.parse.unquote(parsed.hostname).translate(_ascii_lowercase)

    # Deprecated accessors to moved functionality

    @property
    def unresolved_remote(self):
        return self.remote.hostinfo

    @unresolved_remote.setter
    def unresolved_remote(self, value):
        # should get a big fat deprecation warning
        if value is None:
            self.remote = UndecidedRemote('coap', None)
        else:
            self.remote = UndecidedRemote('coap', value)

    @property
    def requested_scheme(self):
        if self.code.is_request():
            return self.remote.scheme
        else:
            return self.request.requested_scheme

    @requested_scheme.setter
    def requested_scheme(self, value):
        self.remote = UndecidedRemote(value, self.remote.hostinfo)

    @property
    def requested_proxy_uri(self):
        return self.request.opt.proxy_uri

    @property
    def requested_hostinfo(self):
        return self.request.opt.uri_host or self.request.unresolved_remote

    @property
    def requested_path(self):
        return self.request.opt.uri_path

    @property
    def requested_query(self):
        return self.request.opt.uri_query

class UndecidedRemote(
        namedtuple("_UndecidedRemote", ("scheme", "hostinfo")),
        interfaces.EndpointAddress
        ):
    """Remote that is set on messages that have not been sent through any any
    transport.

    It describes scheme, hostname and port that were set in
    :meth:`.set_request_uri()` or when setting a URI per Message constructor.

    * :attr:`scheme`: The scheme string
    * :attr:`hostinfo`: The authority component of the URI, as it would occur
      in the URI.
    """

    @classmethod
    def from_pathless_uri(cls, uri: str) -> UndecidedRemote:
        """Create an UndecidedRemote for a given URI that has no query, path,
        fragment or other components not expressed in an UndecidedRemote

        >>> from aiocoap.message import UndecidedRemote
        >>> UndecidedRemote.from_pathless_uri("coap://localhost")
        UndecidedRemote(scheme='coap', hostinfo='localhost')
        >>> UndecidedRemote.from_pathless_uri("coap+tcp://[::1]:1234")
        UndecidedRemote(scheme='coap+tcp', hostinfo='[::1]:1234')
        """

        parsed = urllib.parse.urlparse(uri)

        if parsed.username or parsed.password:
            raise ValueError("User name and password not supported.")

        if parsed.path not in ('', '/') or parsed.query or parsed.fragment:
            raise ValueError("Paths and query and fragment can not be set on an UndecidedRemote")

        return cls(parsed.scheme, parsed.netloc)

_ascii_lowercase = str.maketrans(string.ascii_uppercase, string.ascii_lowercase)

_quote_for_path = quote_factory(unreserved + sub_delims + ':@')
_quote_for_query = quote_factory(unreserved + "".join(c for c in sub_delims if c != '&') + ':@/?')

#: Result that can be returned from a render method instead of a Message when
#: due to defaults (eg. multicast link-format queries) or explicit
#: configuration (eg. the No-Response option), no response should be sent at
#: all. Note that per RFC7967 section 2, an ACK is still sent to a CON
#: request.
#:
#: Depercated; set the no_response option on a regular response instead (see
#: :meth:`.interfaces.Resource.render` for details).
NoResponse = Sentinel("NoResponse")



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Basic implementation of CoAP-CoAP proxying

This is work in progress and not yet part of the API."""

import asyncio
import functools
import ipaddress
import logging
import warnings

from .. import numbers, interfaces, message, error, util, resource
from ..blockwise import Block1Spool, Block2Cache

class CanNotRedirect(error.ConstructionRenderableError):
    message = "Proxy redirection failed"

class NoUriSplitting(CanNotRedirect):
    code = numbers.codes.NOT_IMPLEMENTED
    message = "URI splitting not implemented, please use Proxy-Scheme."

class IncompleteProxyUri(CanNotRedirect):
    code = numbers.codes.BAD_REQUEST
    message = "Proxying requires Proxy-Scheme and Uri-Host"

class NotAForwardProxy(CanNotRedirect):
    code = numbers.codes.PROXYING_NOT_SUPPORTED
    message = "This is a reverse proxy, not a forward one."

class NoSuchHostname(CanNotRedirect):
    code = numbers.codes.NOT_FOUND
    message = ""

class CanNotRedirectBecauseOfUnsafeOptions(CanNotRedirect):
    code = numbers.codes.BAD_OPTION

    def __init__(self, options):
        self.message = "Unsafe options in request: %s" % (", ".join(str(o.number) for o in options))

def raise_unless_safe(request, known_options):
    """Raise a BAD_OPTION CanNotRedirect unless all options in request are
    safe to forward or known"""

    known_options = set(known_options).union({
        # it is expected that every proxy is aware of these options even though
        # one of them often doesn't need touching
        numbers.OptionNumber.URI_HOST,
        numbers.OptionNumber.URI_PORT,
        numbers.OptionNumber.URI_PATH,
        numbers.OptionNumber.URI_QUERY,
        # handled by the Context
        numbers.OptionNumber.BLOCK1,
        numbers.OptionNumber.BLOCK2,
        # handled by the proxy resource
        numbers.OptionNumber.OBSERVE,
        })

    unsafe_options = [o for o in request.opt.option_list() if o.number.is_unsafe() and o.number not in known_options]
    if unsafe_options:
        raise CanNotRedirectBecauseOfUnsafeOptions(unsafe_options)

class Proxy(interfaces.Resource):
    # other than in special cases, we're trying to be transparent wrt blockwise transfers
    interpret_block_options = False

    def __init__(self, outgoing_context, logger=None):
        super().__init__()
        # Provide variables for render_to_pipe
        # FIXME this is copied from aiocoap.resource's __init__ -- but on the
        # long run proxying shouldn't rely on that anyway but implement
        # render_to_pipe right on its own
        self._block1 = Block1Spool()
        self._block2 = Block2Cache()

        self.outgoing_context = outgoing_context
        self.log = logger or logging.getLogger('proxy')

        self._redirectors = []

    def add_redirector(self, redirector):
        self._redirectors.append(redirector)

    def apply_redirection(self, request):
        for r in self._redirectors:
            result = r.apply_redirection(request)
            if result is not None:
                return result
        return None

    async def needs_blockwise_assembly(self, request):
        return self.interpret_block_options

    async def render(self, request):
        # FIXME i'd rather let the application do with the message whatever it
        # wants. everything the responder needs of the request should be
        # extracted beforehand.
        request = request.copy(mid=None, token=None)

        try:
            request = self.apply_redirection(request)
        except CanNotRedirect as e:
            return e.to_message()

        if request is None:
            response = await super().render(request)
            if response is None:
                raise IncompleteProxyUri("No matching proxy rule")
            return response

        try:
            response = await self.outgoing_context.request(request, handle_blockwise=self.interpret_block_options).response
        except error.TimeoutError:
            return message.Message(code=numbers.codes.GATEWAY_TIMEOUT)

        raise_unless_safe(response, ())

        response.mtype = None
        response.mid = None
        response.remote = None
        response.token = None

        return response

    # Not inheriting from them because we do *not* want the .render() in the
    # resolution tree (it can't deal with None requests, which are used among
    # proxy implementations)
    render_to_pipe = resource.Resource.render_to_pipe

class ProxyWithPooledObservations(Proxy, interfaces.ObservableResource):
    def __init__(self, outgoing_context, logger=None):
        super(ProxyWithPooledObservations, self).__init__(outgoing_context, logger)

        self._outgoing_observations = {}

    @staticmethod
    def _cache_key(request):
        return request.get_cache_key([numbers.optionnumbers.OptionNumber.OBSERVE])

    def _peek_observation_for(self, request):
        """Return the augmented request (see _get_obervation_for) towards a
        resource, or raise KeyError"""
        cachekey = self._cache_key(request)

        return self._outgoing_observations[cachekey]

    def _get_observation_for(self, request):
        """Return an existing augmented request towards a resource or create one.

        An augmented request is an observation request that has some additional
        properties (__users, __cachekey, __latest_response), which are used in
        ProxyWithPooledObservations to immediately serve responses from
        observed resources, and to tear the observations down again."""

        # see ProxiedResource.render
        request = request.copy(mid=None, remote=None, token=None)
        request = self.apply_redirection(request)

        cachekey = self._cache_key(request)

        try:
            obs = self._outgoing_observations[cachekey]
        except KeyError:
            obs = self._outgoing_observations[cachekey] = self.outgoing_context.request(request)
            obs.__users = set()
            obs.__cachekey = cachekey
            obs.__latest_response = None # this becomes a cached response right after the .response comes in (so only use this after waiting for it), and gets updated when new responses arrive.

            def when_first_request_done(result, obs=obs):
                obs.__latest_response = result.result()
            obs.response.add_done_callback(when_first_request_done)

            def cb(incoming_message, obs=obs):
                self.log.info("Received incoming message %r, relaying it to %d clients", incoming_message, len(obs.__users))
                obs.__latest_response = incoming_message
                for observationserver in set(obs.__users):
                    observationserver.trigger(incoming_message.copy())
            obs.observation.register_callback(cb)
            def eb(exception, obs=obs):
                if obs.__users:
                    code = numbers.codes.INTERNAL_SERVER_ERROR
                    payload = b""
                    if isinstance(exception, error.RenderableError):
                        code = exception.code
                        payload = exception.message.encode('ascii')
                    self.log.debug("Received error %r, which did not lead to unregistration of the clients. Actively deregistering them with %s %r.", exception, code, payload)
                    for u in list(obs.__users):
                        u.trigger(message.Message(code=code, payload=payload))
                    if obs.__users:
                        self.log.error("Observations survived sending them an error message.")
                else:
                    self.log.debug("Received error %r, but that seems to have been passed on cleanly to the observers as they are gone by now.", exception)
            obs.observation.register_errback(eb)

        return obs

    def _add_observation_user(self, clientobservationrequest, serverobservation):
        clientobservationrequest.__users.add(serverobservation)

    def _remove_observation_user(self, clientobservationrequest, serverobservation):
        clientobservationrequest.__users.remove(serverobservation)
        # give the request that just cancelled time to be dealt with before
        # dropping the __latest_response
        asyncio.get_event_loop().call_soon(self._consider_dropping, clientobservationrequest)

    def _consider_dropping(self, clientobservationrequest):
        if not clientobservationrequest.__users:
            self.log.debug("Last client of observation went away, deregistering with server.")
            self._outgoing_observations.pop(clientobservationrequest.__cachekey)
            if not clientobservationrequest.observation.cancelled:
                clientobservationrequest.observation.cancel()

    async def add_observation(self, request, serverobservation):
        """As ProxiedResource is intended to be just the proxy's interface
        toward the Context, accepting observations is handled here, where the
        observations handling can be defined by the subclasses."""

        try:
            clientobservationrequest = self._get_observation_for(request)
        except CanNotRedirect:
            pass # just don't accept the observation, the rest will be taken care of at rendering
        else:
            self._add_observation_user(clientobservationrequest, serverobservation)
            serverobservation.accept(functools.partial(self._remove_observation_user, clientobservationrequest, serverobservation))

    async def render(self, request):
        # FIXME this is evaulated twice in the implementation (once here, but
        # unless it's an observation what matters is inside the super call),
        # maybe this needs to hook in differently than by subclassing and
        # calling super.
        self.log.info("render called")
        redirected_request = request.copy()

        try:
            redirected_request = self.apply_redirection(redirected_request)
            if redirected_request is None:
                return await super().render(request)
            clientobservationrequest = self._peek_observation_for(redirected_request)
        except (KeyError, CanNotRedirect) as e:
            if not isinstance(e, CanNotRedirect) and request.opt.observe is not None:
                self.log.warning("No matching observation found: request is %r (cache key %r), outgoing observations %r", redirected_request, self._cache_key(redirected_request), self._outgoing_observations)

                return message.Message(code=numbers.codes.BAD_OPTION, payload="Observe option can not be proxied without active observation.".encode('utf8'))
            self.log.debug("Request is not an observation or can't be proxied, passing it on to regular proxying mechanisms.")
            return await super(ProxyWithPooledObservations, self).render(request)
        else:
            self.log.info("Serving request using latest cached response of %r", clientobservationrequest)
            await clientobservationrequest.response
            cached_response = clientobservationrequest.__latest_response
            cached_response.mid = None
            cached_response.token = None
            cached_response.remote = None
            cached_response.mtype = None
            return cached_response


class ForwardProxy(Proxy):
    def apply_redirection(self, request):
        request = request.copy()
        if request.opt.proxy_uri is not None:
            raise NoUriSplitting
        if request.opt.proxy_scheme is None:
            return super().apply_redirection(request)
        if request.opt.uri_host is None:
            raise IncompleteProxyUri

        raise_unless_safe(request, (numbers.OptionNumber.PROXY_SCHEME, numbers.OptionNumber.URI_HOST, numbers.OptionNumber.URI_PORT))

        request.remote = message.UndecidedRemote(request.opt.proxy_scheme, util.hostportjoin(request.opt.uri_host, request.opt.uri_port))
        request.opt.proxy_scheme = None
        request.opt.uri_port = None
        forward_host = request.opt.uri_host
        try:
            # I'd prefer to not do if-by-try, but the ipaddress doesn't seem to
            # offer any other choice
            ipaddress.ip_address(request.opt.uri_host)

            warnings.warn("URI-Host looks like IPv6 but has no square "
                    "brackets. This is deprecated, see "
                    "https://github.com/chrysn/aiocoap/issues/216",
                    DeprecationWarning)
        except ValueError:
            pass
        else:
            request.opt.uri_host = None
        if forward_host.startswith('['):
            # IPv6 or future literals are not recognized by ipaddress which
            # does not look at host-encoded form
            request.opt.uri_host = None

        # Maybe the URI-Host matches a known forwarding -- in that case, catch that.
        redirected = super(ForwardProxy, self).apply_redirection(request)
        if redirected is not None:
            return redirected

        return request

class ForwardProxyWithPooledObservations(ForwardProxy, ProxyWithPooledObservations):
    pass

class ReverseProxy(Proxy):
    def __init__(self, *args, **kwargs):
        import warnings
        warnings.warn("ReverseProxy has become moot due to proxy operation "
                "changes, just instanciate Proxy and set the appropriate "
                "redirectors", DeprecationWarning, stacklevel=1)
        super().__init__(*args, **kwargs)

class ReverseProxyWithPooledObservations(ReverseProxy, ProxyWithPooledObservations):
    pass

class Redirector():
    def apply_redirection(self, request):
        return None

class NameBasedVirtualHost(Redirector):
    def __init__(self, match_name, target, rewrite_uri_host=False, use_as_proxy=False):
        self.match_name = match_name
        self.target = target
        self.rewrite_uri_host = rewrite_uri_host
        self.use_as_proxy = use_as_proxy

    def apply_redirection(self, request):
        raise_unless_safe(request, ())

        if self._matches(request.opt.uri_host):
            if self.use_as_proxy:
                request.opt.proxy_scheme = request.remote.scheme
            if self.rewrite_uri_host:
                request.opt.uri_host, _ = util.hostportsplit(self.target)
            request.unresolved_remote = self.target
            return request

    def _matches(self, hostname):
        return hostname == self.match_name

class SubdomainVirtualHost(NameBasedVirtualHost):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.rewrite_uri_host:
            raise TypeError("rewrite_uri_host makes no sense with subdomain virtual hosting")

    def _matches(self, hostname):
        return hostname.endswith('.' + self.match_name)

class UnconditionalRedirector(Redirector):
    def __init__(self, target, use_as_proxy=False):
        self.target = target
        self.use_as_proxy = use_as_proxy

    def apply_redirection(self, request):
        raise_unless_safe(request, ())

        if self.use_as_proxy:
            request.opt.proxy_scheme = request.remote.scheme
        request.unresolved_remote = self.target
        return request

class SubresourceVirtualHost(Redirector):
    def __init__(self, path, target):
        self.path = tuple(path)
        self.target = target

    def apply_redirection(self, request):
        raise_unless_safe(request, ())

        if self.path == request.opt.uri_path[:len(self.path)]:
            request.opt.uri_path = request.opt.uri_path[len(self.path):]
            request.unresolved_remote = self.target
            return request



# SPDX-FileCopyrightText: Michael Burrows <mjb@asplake.co.uk>
#
# SPDX-License-Identifier: BSD-3-Clause

'''
Parse and format link headers according to RFC 5988 "Web Linking".

Usage (assuming a suitable headers object in the environment):

>>> headers['Link'] = str(LinkHeader([Link("http://example.com/foo", rel="self"),
...                                   Link("http://example.com", rel="up")]))
>>> headers['Link']
'<http://example.com/foo>; rel=self, <http://example.com>; rel=up'
>>> parse(headers['Link'])
LinkHeader([Link('http://example.com/foo', rel='self'), Link('http://example.com', rel='up')])

Blank and missing values roundtrip correctly:

>>> format_link(parse('</s/1>; obs; if="core.s"; foo=""'))
'<</s/1>; obs; if=core.s; foo="">'

Conversions to and from json-friendly list-based structures are also provided:

>>> parse(headers['Link']).to_py()
[['http://example.com/foo', [['rel', 'self']]], ['http://example.com', [['rel', 'up']]]]
>>> str(LinkHeader([['http://example.com/foo', [['rel', 'self']]],
...                 ['http://example.com', [['rel', 'up']]]]))
'<http://example.com/foo>; rel=self, <http://example.com>; rel=up'

For further information see parse(), LinkHeader and Link.
'''

import re
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

__all__ = ['parse', 'format_links', 'format_link', 'LinkHeader', 'Link', 'ParseException']

SINGLE_VALUED_ATTRS = ['rel', 'anchor', 'rev', 'media', 'title', 'title*', 'type']

#
# Regexes for link header parsing.  TOKEN and QUOTED in particular should conform to RFC2616.
#
# Acknowledgement: The QUOTED regexp is based on
# http://stackoverflow.com/questions/249791/regexp-for-quoted-string-with-escaping-quotes/249937#249937
#
# Trailing spaces are consumed by each pattern.  The RE_HREF pattern also allows for any leading spaces.
#

QUOTED        = r'"((?:[^"\\]|\\.)*)"'                  # double-quoted strings with backslash-escaped double quotes
TOKEN         = r'([^()<>@,;:\"\[\]?={}\s]+)'           # non-empty sequence of non-separator characters
RE_COMMA_HREF = re.compile(r' *,? *< *([^>]*) *> *')    # includes ',' separator; no attempt to check URI validity
RE_ONLY_TOKEN = re.compile(r'^%(TOKEN)s$' % locals())
RE_ATTR       = re.compile(r'%(TOKEN)s *(?:= *(%(TOKEN)s|%(QUOTED)s))? *' % locals())
RE_SEMI       = re.compile(r'; *')
RE_COMMA      = re.compile(r', *')


def parse(header):
    '''Parse a link header string, returning a LinkHeader object:

    >>> parse('<http://example.com/foo>; rel="foo bar", <http://example.com>; rel=up; type=text/html')
    LinkHeader([Link('http://example.com/foo', rel='foo bar'), Link('http://example.com', rel='up', type='text/html')])

    ParseException is raised in the event that the input string is not parsed completely:

    >>> parse('<http://example.com/foo> error') #doctest: +SKIP
    Traceback (most recent call last):
        ...
    ParseException: ('link_header.parse() failed near %s', "'error'")
    '''
    scanner = _Scanner(header)
    links = []
    while scanner.scan(RE_COMMA_HREF):
        href = scanner[1]
        attrs = []
        while scanner.scan(RE_SEMI):
            if scanner.scan(RE_ATTR):
                attr_name, token, quoted = scanner[1], scanner[3], scanner[4]
                if quoted is not None:
                    attrs.append([attr_name, quoted.replace(r'\"', '"')])
                elif token is not None:
                    attrs.append([attr_name, token])
                else:
                    attrs.append([attr_name, None])
        links.append(Link(href, attrs))

    if scanner.buf:
        raise ParseException("link_header.parse() failed near %s", repr(scanner.buf))

    return LinkHeader(links)

def format_links(*args, **kwargs):
    return str(LinkHeader(*args, **kwargs))

def format_link(*args, **kwargs):
    return str(Link(*args, **kwargs))


class ParseException(Exception):
    pass


class LinkHeader(object):
    '''Represents a sequence of links that can be formatted together as a link header.
    '''
    def __init__(self, links=None):
        '''Initializes a LinkHeader object with a list of Link objects or with
        list of parameters from which Link objects can be created:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])
        LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])
        >>> LinkHeader([['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]])
        LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')])

        The Link objects can be accessed afterwards via the `links` property.

        String conversion follows the spec:

        >>> str(LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]))
        '<http://example.com/foo>; rel=foo, <http://example.com>; rel=up'

        Conversion to json-friendly list-based structures:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]).to_py()
        [['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]]

        '''

        self.links = [
            link if isinstance(link, Link) else Link(*link)
            for link in links or []]

    def to_py(self):
        '''Supports list conversion:

        >>> LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]).to_py()
        [['http://example.com/foo', [['rel', 'foo']]], ['http://example.com', [['rel', 'up']]]]
        '''
        return [link.to_py() for link in self.links]

    def __repr__(self):
        return 'LinkHeader([%s])' % ', '.join(repr(link) for link in self.links)

    def __str__(self):
        '''Formats a link header:

        >>> str(LinkHeader([Link('http://example.com/foo', rel='foo'), Link('http://example.com', rel='up')]))
        '<http://example.com/foo>; rel=foo, <http://example.com>; rel=up'
        '''
        return ', '.join(str(link) for link in self.links)

    def links_by_attr_pairs(self, pairs):
        '''Lists links that have attribute pairs matching all the supplied pairs:

         >>> parse('<http://example.com/foo>; rel="foo", <http://example.com>; rel="up"'
         ...      ).links_by_attr_pairs([('rel', 'up')])
         [Link('http://example.com', rel='up')]
        '''
        return [link
                for link in self.links
                if all([key, value] in link.attr_pairs
                       for key, value in pairs)]

class Link(object):
    '''Represents a single link.
    '''

    def __init__(self, href, attr_pairs=None, **kwargs):
        '''Initializes a Link object with an href and attributes either in
        the form of a sequence of key/value pairs &/or as keyword arguments.
        The sequence form allows to be repeated.  Attributes may be accessed
        subsequently via the `attr_pairs` property.

        String conversion follows the spec:

        >>> str(Link('http://example.com', [('foo', 'bar'), ('foo', 'baz')], rel='self'))
        '<http://example.com>; foo=bar; foo=baz; rel=self'

        Conversion to json-friendly list-based structures:

        >>> Link('http://example.com', [('foo', 'bar'), ('foo', 'baz')], rel='self').to_py()
        ['http://example.com', [['foo', 'bar'], ['foo', 'baz'], ['rel', 'self']]]
        '''
        self.href = href
        self.attr_pairs = [
            list(pair)
            for pair in (attr_pairs or []) + list(kwargs.items())]

    def to_py(self):
        '''Convert to a json-friendly list-based structure:

        >>> Link('http://example.com', rel='foo').to_py()
        ['http://example.com', [['rel', 'foo']]]
        '''
        return [self.href, self.attr_pairs]

    def __repr__(self):
        '''
        >>> Link('http://example.com', rel='self')
        Link('http://example.com', rel='self')
        '''
        return 'Link(%s)' % ', '.join(
            [
                repr(self.href)
            ] + [
                "%s=%s" % (pair[0], repr(pair[1]))
                for pair in self.attr_pairs])

    def __str__(self):
        '''Formats a single link:

        >>> str(Link('http://example.com/foo', [['rel', 'self']]))
        '<http://example.com/foo>; rel=self'
        >>> str(Link('http://example.com/foo', [['rel', '"quoted"'], ['type', 'text/html'], ['title*', "UTF-8'en'%e2%82%ac%20rates"]]))
        '<http://example.com/foo>; rel="\\\\"quoted\\\\""; type=text/html; title*=UTF-8\\'en\\'%e2%82%ac%20rates'

        Note that there is no explicit support for the title* attribute other
        than to output it unquoted.  Where used, it is up to client applications to
        provide values that meet RFC2231 Section 7.
        '''
        def str_pair(key, value):
            if value is None:
                return key
            elif RE_ONLY_TOKEN.match(value) or key.endswith('*'):
                return '%s=%s' % (key, value)
            else:
                return '%s="%s"' % (key, value.replace('"', r'\"'))
        return '; '.join(['<%s>' % self.href] +
                         [str_pair(key, value)
                          for key, value in self.attr_pairs])

    def __getattr__(self, name):
        '''
        >>> Link('/', rel='self').rel
        'self'
        >>> Link('/', hreflang='EN').hreflang
        ['EN']
        >>> Link('/', foo='bar').foo
        ['bar']
        >>> Link('/', [('foo', 'bar'), ('foo', 'baz')]).foo
        ['bar', 'baz']
        >>> Link('/').rel #doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        AttributeError: No attribute of type 'rel' present
        >>> Link('/').hreflang
        []
        >>> Link('/').foo
        []
        '''
        name_lower = name.lower()
        values = [value
                  for key, value in self.attr_pairs
                  if key.lower() == name_lower]
        if name in SINGLE_VALUED_ATTRS:
            if values:
                return values[0]
            else:
                raise AttributeError("No attribute of type %r present"%name_lower)
        return values

    def __contains__(self, name):
        '''
        >>> 'rel' in Link('/', rel='self')
        True
        >>> 'obs' in Link('/', obs=None)
        True
        >>> 'rel' in Link('/')
        False
        '''
        name_lower = name.lower()
        return any(key.lower() == name_lower for key, value in self.attr_pairs)

    def get_context(self, requested_resource_address):
        '''Return the absolute URI of the context of a link. This is usually
        equals the base address the statement is about (eg. the requested URL
        if the link header was served in a successful HTTP GET request), but
        can be overridden by the anchor parameter.

        >>> Link('../', rel='index').get_context('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/chapter1/'
        >>> Link('', rel='next', anchor='../').get_context('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/'
        '''
        if 'anchor' in self:
            return urljoin(requested_resource_address, self.anchor)
        return requested_resource_address

    def get_target(self, requested_resource_address):
        '''Return the absolute URI of the target of a link. It is determined by
        joining the address from which the link header was retrieved with the
        link-value (inside angular brackets) according to RFC3986 section 5.

        >>> Link('../', rel='index').get_target('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/'
        >>> Link('', rel='next', anchor='../').get_target('http://www.example.com/book1/chapter1/')
        'http://www.example.com/book1/chapter1/'
        '''
        return urljoin(requested_resource_address, self.href)

class _Scanner(object):
    def __init__(self, buf):
        self.buf = buf
        self.match = None

    def __getitem__(self, key):
        return self.match.group(key)

    def scan(self, pattern):
        self.match = pattern.match(self.buf)
        if self.match:
            self.buf = self.buf[self.match.end():]
        return self.match


# For doctest
headers = dict()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains in-place modifications to the LinkHeader module to
satisfy RFC6690 constraints.

It is a general nursery for what aiocoap needs of link-format management before
any of this is split out into its own package.
"""

from .vendored import link_header

class LinkFormat(link_header.LinkHeader):
    def __str__(self):
        return ','.join(str(link) for link in self.links)

class Link(link_header.Link):
    # This is copy-pasted from the link_header module's code, just replacing
    # the '; ' with ';'.
    #
    # Original copyright Michael Burrows <mjb@asplake.co.uk>, distributed under
    # the BSD license
    def __str__(self):
        def str_pair(key, value):
            if value is None:
                return key
# workaround to accomodate copper
#            elif RE_ONLY_TOKEN.match(value) or key.endswith('*'):
#                return '%s=%s' % (key, value)
            else:
                return '%s="%s"' % (key, value.replace('"', r'\"'))
        return ';'.join(['<%s>' % self.href] +
                         [str_pair(key, value)
                          for key, value in self.attr_pairs])

def parse(linkformat):
    data = link_header.parse(linkformat)
    data.__class__ = LinkFormat
    for l in data.links:
        l.__class__ = Link
    return data



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Helper module around getaddrinfo shortcomings"""

import socket
import itertools
import errno

async def getaddrinfo_routechecked(loop, log, host, port):
    """Variant of getaddrinfo that works like AI_ADDRCONFIG should probably work.

    This is not only a workaround for 12377_, but goes beyond it by checking
    routability. Even with 12377 fixed, AI_ADDRCONFIG would just avoid sending
    AAAA requests (if no v6 addresses were available at all). It would still
    not avoid filtering out responses that are just not routable -- the typical
    example being that a host may be in a network with a ULA, so it needs to
    send AAAA requests because they might resolve to the ULA, but if a global
    address comes back, we still don't want that address passed to the
    application.

    The family and flags arguments are not accepted, because they are what is
    being altered (and it'd be hard to tell which combinations would work). It
    roughly behaves as if flags were set to AI_ADDRCONFIG | AI_V4MAPPED, and
    family to AF_INET6. The type and protocol are fixed to SOCK_DGRAM /
    IPPROTO_UDP, because only there we known that connect is side-effect free.

    This function also differs from getaddrinfo in that it returns an
    asynchronous iterator (we don't want to check addresses we don't care
    about), and in that it only returns the sockaddr (because the rest is fixed
    anyway).

    .. _12377: https://sourceware.org/bugzilla/show_bug.cgi?id=12377
    """

    # As an implementation note, we can't go through the AI_V4MAPPED feature,
    # because if we used that, we'd only get V4 responses if there are no V6
    # addresses. As we deliberately want to produce results on V4 addresses
    # when both are provided (conditional on the V6 ones not being routable),
    # we need to convert -- and then the quick way to do things is to use
    # AF_UNSPEC and distinguish later.

    addrinfo = await loop.getaddrinfo(
        host,
        port,
        family=socket.AF_UNSPEC,
        type=socket.SOCK_DGRAM,
        proto=socket.IPPROTO_UDP,
        # Still setting that -- it spares us the traffic of pointless requests.
        flags=socket.AI_ADDRCONFIG,
        )

    if any(a[0] not in (socket.AF_INET6, socket.AF_INET) for a in addrinfo):
        log.warning("Addresses outside of INET and INET6 families ignored.")

    v6_addresses = (sockaddr for (family, *_, sockaddr) in addrinfo if family == socket.AF_INET6)
    # Dress them just as AI_V4MAPPED would do. Note that we can't do (ip, port)
    # instead of sockaddr b/c it's destructured already for the family check,
    # and at that time it may be a different AF with more members.
    v4_addresses = (('::ffff:' + sockaddr[0], sockaddr[1], 0, 0) for (family, *_, sockaddr) in addrinfo if family == socket.AF_INET)

    yielded = 0
    for (ip, port, flowinfo, scope_id) in itertools.chain(v6_addresses, v4_addresses):
        # Side-step connectivity test when explicitly giving an address. This
        # should be purely a cosmetic change, but given we try not to give IP
        # addresses special treatment over host names, it needs to be done
        # somewhere if we don't want connection attempts to explicit V6
        # addresses to err with "No address information found" on V4-only
        # hosts.
        if ip != host:
            with socket.socket(family=socket.AF_INET6, type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP) as tempsock:
                try:
                    tempsock.connect((ip, port))
                except OSError as e:
                    if e.errno == errno.ENETUNREACH:
                        continue

        yield (ip, port, flowinfo, scope_id)
        yielded += 1

    if not yielded:
        # That's the behavior of getaddrinfo -- not return empty (so we
        # shouldn't return empty just b/c everything was filtered either). Do
        # we need to be any more specific? The gaierror are caught later
        # anyway...
        raise socket.gaierror



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains numeric constants that would be expected in the socket
module, but are not exposed there.

This gathers both socket numbers that can be present in the socket module (eg.
the PKTINFO constants) but are not in some versions (eg. on macOS before
<https://bugs.python.org/issue35569> is fixed) and platform dependent constants
that are not generally available at all (the ERR constants).

Where available, the CPython-private IN module is used to obtain some platform
specific constants.

Any hints on where to get them from in a more reliable way are appreciated;
possible options are parsing C header files (at build time?) or interacting
with shared libraries for obtaining the symbols. The right way would probably
be including them in Python in a "other constants defined on this platform for
sockets" module or dictionary.
"""

import sys

try:
    from socket import IPV6_PKTINFO, IPV6_RECVPKTINFO
except ImportError:
    if sys.platform == 'linux':
        # Not sure if here are any Linux builds at all where this is
        # unavailable
        IPV6_PKTINFO = 50
        IPv6_RECVPKTINFO = 49
    elif sys.platform == 'darwin':
        # when __APPLE_USE_RFC_3542 is defined / as would be when
        # https://bugs.python.org/issue35569 is fixed
        IPV6_PKTINFO = 46
        IPV6_RECVPKTINFO = 61
    # Not attempting to make any guesses for other platforms; the udp6 module
    # will fail to import where it needs the specifics

try:
    from IN import IPV6_RECVERR, IP_RECVERR
except ImportError:
    if sys.platform == 'linux':
        IPV6_RECVERR = 25
        IP_RECVERR = 11

# for https://bitbucket.org/pypy/pypy/issues/2648/
try:
    from socket import MSG_ERRQUEUE
except ImportError:
    if sys.platform == 'linux':
        MSG_ERRQUEUE = 8192

HAS_RECVERR = 'IP_RECVERR' in locals() and 'MSG_ERRQUEUE' in locals()
"""Indicates whether the discovered constants indicate that the Linux
`setsockopt(IPV6, RECVERR)` / `recvmsg(..., MSG_ERRQUEUE)` mechanism is
available"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module provides interface base classes to various aiocoap software
components, especially with respect to request and response handling. It
describes `abstract base classes`_ for messages, endpoints etc.

It is *completely unrelated* to the concept of "network interfaces".

.. _`abstract base classes`: https://docs.python.org/3/library/abc"""

from __future__ import annotations

import abc
import asyncio
import warnings

from aiocoap.pipe import Pipe
from aiocoap.numbers.constants import MAX_REGULAR_BLOCK_SIZE_EXP

from typing import Optional, Callable

class MessageInterface(metaclass=abc.ABCMeta):
    """A MessageInterface is an object that can exchange addressed messages over
    unreliable transports. Implementations send and receive messages with
    message type and message ID, and are driven by a Context that deals with
    retransmission.

    Usually, an MessageInterface refers to something like a local socket, and
    send messages to different remote endpoints depending on the message's
    addresses. Just as well, a MessageInterface can be useful for one single
    address only, or use various local addresses depending on the remote
    address.
    """

    @abc.abstractmethod
    async def shutdown(self):
        """Deactivate the complete transport, usually irrevertably. When the
        coroutine returns, the object must have made sure that it can be
        destructed by means of ref-counting or a garbage collector run."""

    @abc.abstractmethod
    def send(self, message):
        """Send a given :class:`Message` object"""

    @abc.abstractmethod
    async def determine_remote(self, message):
        """Return a value suitable for the message's remote property based on
        its .opt.uri_host or .unresolved_remote.

        May return None, which indicates that the MessageInterface can not
        transport the message (typically because it is of the wrong scheme)."""

class EndpointAddress(metaclass=abc.ABCMeta):
    """An address that is suitable for routing through the application to a
    remote endpoint.

    Depending on the MessageInterface implementation used, an EndpointAddress
    property of a message can mean the message is exchanged "with
    [2001:db8::2:1]:5683, while my local address was [2001:db8:1::1]:5683"
    (typical of UDP6), "over the connected <Socket at
    0x1234>, whereever that's connected to" (simple6 or TCP) or "with
    participant 0x01 of the OSCAP key 0x..., routed over <another
    EndpointAddress>".

    EndpointAddresses are only concstructed by MessageInterface objects,
    either for incoming messages or when populating a message's .remote in
    :meth:`MessageInterface.determine_remote`.

    There is no requirement that those address are always identical for a given
    address. However, incoming addresses must be hashable and hash-compare
    identically to requests from the same context. The "same context", for the
    purpose of EndpointAddresses, means that the message must be eligible for
    request/response, blockwise (de)composition and observations. (For example,
    in a DTLS context, the hash must change between epochs due to RFC7252
    Section 9.1.2).

    So far, it is required that hash-identical objects also compare the same.
    That requirement might go away in future to allow equality to reflect finer
    details that are not hashed. (The only property that is currently known not
    to be hashed is the local address in UDP6, because that is *unknown* in
    initially sent packages, and thus disregarded for comparison but needed to
    round-trip through responses.)
    """

    @property
    @abc.abstractmethod
    def hostinfo(self):
        """The authority component of URIs that this endpoint represents when
        request are sent to it

        Note that the presence of a hostinfo does not necessarily mean that
        globally meaningful or even syntactically valid URI can be constructed
        out of it; use the :attr:`.uri` property for this."""

    @property
    @abc.abstractmethod
    def hostinfo_local(self):
        """The authority component of URIs that this endpoint represents when
        requests are sent from it.

        As with :attr:`.hostinfo`, this does not necessarily produce sufficient
        input for a URI; use :attr:`.uri_local` instead."""

    @property
    def uri(self):
        """Deprecated alias for uri_base"""
        return self.uri_base

    @property
    @abc.abstractmethod
    def uri_base(self):
        """The base URI for the peer (typically scheme plus .hostinfo).

        This raises :class:`.error.AnonymousHost` when executed on an address
        whose peer coordinates can not be expressed meaningfully in a URI."""

    @property
    @abc.abstractmethod
    def uri_base_local(self):
        """The base URI for the local side of this remote.

        This raises :class:`.error.AnonymousHost` when executed on an address
        whose local coordinates can not be expressed meaningfully in a URI."""

    @property
    @abc.abstractmethod
    def is_multicast(self):
        """True if the remote address is a multicast address, otherwise false."""

    @property
    @abc.abstractmethod
    def is_multicast_locally(self):
        """True if the local address is a multicast address, otherwise false."""

    @property
    @abc.abstractmethod
    def scheme(Self):
        """The that is used with addresses of this kind

        This is usually a class property. It is applicable to both sides of the
        communication. (Should there ever be a scheme that addresses the
        participants differently, a scheme_local will be added.)"""

    maximum_block_size_exp = MAX_REGULAR_BLOCK_SIZE_EXP
    """The maximum negotiated block size that can be sent to this remote."""

    # Giving some slack so that barely-larger messages (like OSCORE typically
    # are) don't get fragmented -- but still for migration to maximum message
    # size so we don't have to guess any more how much may be option and how
    # much payload
    maximum_payload_size = 1124
    """The maximum payload size that can be sent to this remote. Only relevant
    if maximum_block_size_exp is 7. This will be removed in favor of a maximum
    message size when the block handlers can get serialization length
    predictions from the remote."""

    def as_response_address(self):
        """Address to be assigned to a response to messages that arrived with
        this message

        This can (and does, by default) return self, but gives the protocol the
        opportunity to react to create a modified copy to deal with variations
        from multicast.
        """
        return self

    @property
    def authenticated_claims(self):
        """Iterable of objects representing any claims (e.g. an identity, or
        generally objects that can be used to authorize particular accesses)
        that were authenticated for this remote.

        This is experimental and may be changed without notice.

        Its primary use is on the server side; there, a request handler (or
        resource decorator) can use the claims to decide whether the client is
        authorized for a particular request. Use on the client side is planned
        as a requirement on a request, although (especially on side-effect free
        non-confidential requests) it can also be used in response
        processing."""
        # "no claims" is a good default
        return ()

    @property
    @abc.abstractmethod
    def blockwise_key(self):
        """A hashable (ideally, immutable) value that is only the same for
        remotes from which blocks may be combined. (With all current transports
        that means that the network addresses need to be in there, and the
        identity of the security context).

        It does *not* just hinge on the identity of the address object, as a
        first block may come in an OSCORE group request and follow-ups may come
        in pairwise requests. (And there might be allowed relaxations on the
        transport under OSCORE, but that'd need further discussion)."""
        # FIXME: should this behave like something that keeps the address
        # alive? Conversely, if the address gets deleted, can this reach the
        # block keys and make their stuff vanish from the caches?
        #
        # FIXME: what do security mechanisms best put here? Currently it's a
        # wild mix of keys (OSCORE -- only thing guaranteed to never be reused;
        # DTLS client because it's available) and claims (DTLS server, because
        # it's available and if the claims set matches it can't be that wrong
        # either can it?)

class MessageManager(metaclass=abc.ABCMeta):
    """The interface an entity that drives a MessageInterface provides towards
    the MessageInterface for callbacks and object acquisition."""

    @abc.abstractmethod
    def dispatch_message(self, message):
        """Callback to be invoked with an incoming message"""

    @abc.abstractmethod
    def dispatch_error(self, error: Exception, remote):
        """Callback to be invoked when the operating system indicated an error
        condition from a particular remote."""

    @property
    @abc.abstractmethod
    def client_credentials(self):
        """A CredentialsMap that transports should consult when trying to
        establish a security context"""

class TokenInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def send_message(self, message, messageerror_monitor) -> Optional[Callable[[], None]]:
        """Send a message. If it returns a a callable, the caller is asked to
        call in case it no longer needs the message sent, and to dispose of if
        it doesn't intend to any more.

        messageerror_monitor is a function that will be called at most once by
        the token interface: When the underlying layer is indicating that this
        concrete message could not be processed. This is typically the case for
        RSTs on from the message layer, and used to cancel observations. Errors
        that are not likely to be specific to a message (like retransmission
        timeouts, or ICMP errors) are reported through dispatch_error instead.
        (While the information which concrete message triggered that might be
        available, it is not likely to be relevant).

        Currently, it is up to the TokenInterface to unset the no_response
        option in response messages, and to possibly not send them."""

    @abc.abstractmethod
    async def fill_or_recognize_remote(self, message):
        """Return True if the message is recognized to already have a .remote
        managedy by this TokenInterface, or return True and set a .remote on
        message if it should (by its unresolved remote or Uri-* options) be
        routed through this TokenInterface, or return False otherwise."""

class TokenManager(metaclass=abc.ABCMeta):
    # to be described in full; at least there is a dispatch_error in analogy to MessageManager's
    pass

class RequestInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def fill_or_recognize_remote(self, message):
        pass

    @abc.abstractmethod
    def request(self, request: Pipe):
        pass

class RequestProvider(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def request(self, request_message):
        """Create and act on a a :class:`Request` object that will be handled
        according to the provider's implementation.

        Note that the request is not necessarily sent on the wire immediately;
        it may (but, depend on the transport does not necessarily) rely on the
        response to be waited for."""

class Request(metaclass=abc.ABCMeta):
    """A CoAP request, initiated by sending a message. Typically, this is not
    instanciated directly, but generated by a :meth:`RequestProvider.request`
    method."""

    response = """A future that is present from the creation of the object and \
        fullfilled with the response message.

        When legitimate errors occur, this becomes an aiocoap.Error. (Eg. on
        any kind of network failure, encryption trouble, or protocol
        violations). Any other kind of exception raised from this is a bug in
        aiocoap, and should better stop the whole application.
        """

class Resource(metaclass=abc.ABCMeta):
    """Interface that is expected by a :class:`.protocol.Context` to be present
    on the serversite, which renders all requests to that context."""

    def __init__(self):
        super().__init__()

        # FIXME: These keep addresses alive, and thus possibly transports.
        # Going through the shutdown dance per resource seems extraneous.
        # Options are to accept addresses staying around (making sure they
        # don't keep their transports alive, if that's a good idea), to hash
        # them, or to make them weak.

        from .blockwise import Block1Spool, Block2Cache
        self._block1 = Block1Spool()
        self._block2 = Block2Cache()

    @abc.abstractmethod
    async def render(self, request):
        """Return a message that can be sent back to the requester.

        This does not need to set any low-level message options like remote,
        token or message type; it does however need to set a response code.

        A response returned may carry a no_response option (which is actually
        specified to apply to requests only); the underlying transports will
        decide based on that and its code whether to actually transmit the
        response."""

    @abc.abstractmethod
    async def needs_blockwise_assembly(self, request):
        """Indicator to the :class:`.protocol.Responder` about whether it
        should assemble request blocks to a single request and extract the
        requested blocks from a complete-resource answer (True), or whether
        the resource will do that by itself (False)."""

    async def _render_to_pipe(self, request: Pipe):
        if not hasattr(self, "_block1"):
            warnings.warn("No attribute _block1 found on instance of "
                    f"{type(self).__name__}, make sure its __init__ code "
                    "properly calls super()!", DeprecationWarning)

            from .blockwise import Block1Spool, Block2Cache
            self._block1 = Block1Spool()
            self._block2 = Block2Cache()

        req = request.request

        if await self.needs_blockwise_assembly(req):
            req = self._block1.feed_and_take(req)

            # Note that unless the lambda get's called, we're not fully
            # accessing req any more -- we're just looking at its block2
            # option, and the blockwise key extracted earlier.
            res = await self._block2.extract_or_insert(req, lambda: self.render(req))

            res.opt.block1 = req.opt.block1
        else:
            res = await self.render(req)

        request.add_response(res, is_last=True)

    async def render_to_pipe(self, request: Pipe):
        """Create any number of responses (as indicated by the request) into
        the request stream.

        This method is provided by the base Resource classes; if it is
        overridden, then :meth:`~.Resource.render`, :meth:`needs_blockwise_assembly` and
        :meth:`~.ObservableResource.add_observation` are not used any more.
        (They still need to be implemented to comply with the interface
        definition, which is yet to be updated)."""
        warnings.warn("Request interface is changing: Resources should "
                "implement render_to_pipe or inherit from "
                "resource.Resource which implements that based on any "
                "provided render methods", DeprecationWarning)
        if isinstance(self, ObservableResource):
            # While the above deprecation is used, a resource previously
            # inheriting from (X, ObservableResource) with X inheriting from
            # Resource might find itself using this method. When migrating over
            # to inheriting from resource.Resource, this error will become
            # apparent and this can die with the rest of this workaround.
            return await ObservableResource._render_to_pipe(self, request)
        return await self._render_to_pipe(request)

class ObservableResource(Resource, metaclass=abc.ABCMeta):
    """Interface the :class:`.protocol.ServerObservation` uses to negotiate
    whether an observation can be established based on a request.

    This adds only functionality for registering and unregistering observations;
    the notification contents will be retrieved from the resource using the
    regular :meth:`~.Resource.render` method from crafted (fake) requests.
    """
    @abc.abstractmethod
    async def add_observation(self, request, serverobservation):
        """Before the incoming request is sent to :meth:`~.Resource.render`, the
        :meth:`.add_observation` method is called. If the resource chooses to
        accept the observation, it has to call the
        `serverobservation.accept(cb)` with a callback that will be called when
        the observation ends. After accepting, the ObservableResource should
        call `serverobservation.trigger()` whenever it changes its state; the
        ServerObservation will then initiate notifications by having the
        request rendered again."""


    async def _render_to_pipe(self, pipe):
        from .protocol import ServerObservation

        # If block2:>0 comes along, we'd just ignore the observe
        if pipe.request.opt.observe != 0:
            return await Resource._render_to_pipe(self, pipe)

        # If block1 happens here, we can probably just not support it for the
        # time being. (Given that block1 + observe is untested and thus does
        # not work so far anyway).

        servobs = ServerObservation()
        await self.add_observation(pipe.request, servobs)

        try:
            first_response = await self.render(pipe.request)

            if not servobs._accepted or servobs._early_deregister or \
                    not first_response.code.is_successful():
                pipe.add_response(first_response, is_last=True)
                return

            # FIXME: observation numbers should actually not be per
            # asyncio.task, but per (remote, token). if a client renews an
            # observation (possibly with a new ETag or whatever is deemed
            # legal), the new observation events should still carry larger
            # numbers. (if they did not, the client might be tempted to discard
            # them).
            first_response.opt.observe = next_observation_number = 0
            # If block2 were to happen here, we'd store the full response
            # here, and pick out block2:0.
            pipe.add_response(first_response, is_last=False)

            while True:
                await servobs._trigger
                # if you wonder why the lines around this are not just `response =
                # await servobs._trigger`, have a look at the 'double' tests in
                # test_observe.py: A later triggering could have replaced
                # servobs._trigger in the meantime.
                response = servobs._trigger.result()
                servobs._trigger = asyncio.get_running_loop().create_future()

                if response is None:
                    response = await self.render(pipe.request)

                # If block2 were to happen here, we'd store the full response
                # here, and pick out block2:0.

                is_last = servobs._late_deregister or not response.code.is_successful()
                if not is_last:
                    next_observation_number += 1
                    response.opt.observe = next_observation_number

                pipe.add_response(response, is_last=is_last)

                if is_last:
                    return
        finally:
            servobs._cancellation_callback()

    async def render_to_pipe(self, request: Pipe):
        warnings.warn("Request interface is changing: Resources should "
                "implement render_to_pipe or inherit from "
                "resource.Resource which implements that based on any "
                "provided render methods", DeprecationWarning)
        return await self._render_to_pipe(request)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Implementations of draft-ietf-core-resource-directory

Currently, this only contains the :mod:`client` module. There is an
implementation of the server side contained in aiocoap, but it is currently
accessible only as :mod:`command line application <aiocoap.cli.rd>`.
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Module in which all meaningful numbers are collected. Most of the submodules
correspond to IANA registries.

The contents of the :mod:`.constants`, :mod:`.types` and :mod:`.codes` modules
are accessible through this module directly; :mod:`.contentformat`'s and
:mod:`.optionnumbers`' sole :class:`.contentformat.ContentFormat` and
:class:`.optionnumbers.OptionNumber` classes are accessible in the same way.
"""

import warnings
import string

from . import constants, types, codes
# flake8 doesn't see through the global re-export
from .constants import * # noqa: F401 F403
from .types import * # noqa: F401 F403
from .codes import * # noqa: F401 F403
from .contentformat import ContentFormat, _MediaTypes, _MediaTypesRev
from .optionnumbers import OptionNumber

__all__ = constants.__all__ + types.__all__ + codes.__all__ + ['OptionNumber', 'ContentFormat']

media_types = _MediaTypes()
media_types_rev = _MediaTypesRev()

def __getattr__(name):
    if name[0] in string.ascii_uppercase and hasattr(constants._default_transport_tuning, name):
        warnings.warn(f"{name} is deprecated, use through the message's transport_tuning instead", DeprecationWarning, stacklevel=2)
        return getattr(constants._default_transport_tuning, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""List of known values for the CoAP "Type" field.

As this field is only 2 bits, its valid values are comprehensively enumerated
in the `Type` object.
"""

from enum import IntEnum

class Type(IntEnum):
    CON = 0 # Confirmable
    NON = 1 # Non-confirmable
    ACK = 2 # Acknowledgement
    RST = 3 # Reset

    def __str__(self):
        return self.name

CON, NON, ACK, RST = Type.CON, Type.NON, Type.ACK, Type.RST

__all__ = ['Type', 'CON', 'NON', 'ACK', 'RST']



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Extensions to asyncio and workarounds around its shortcomings"""

import sys

def py38args(**kwargs):
    """Wrapper around kwargs that replaces them with an empty list for Python
    versions earlier than 3.8.

    This is used to assign a name in asyncio.create_task to pass in a name."""
    if sys.version_info >= (3, 8):
        return kwargs
    else:
        return {}



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This helper script can be used to easily inspect aiocoap's environment
autodetection (ie. whether all modules required for particular subsystems are
available, losely corresponding to the "features" made available through
setup.py); run it as `python3 -m aiocoap.cli.defaults`."""

import sys
from aiocoap.meta import version
from aiocoap.defaults import has_reuse_port, get_default_clienttransports, \
    get_default_servertransports, missing_module_functions
import argparse
import os

if __name__ == "__main__":
    p = argparse.ArgumentParser(description=__doc__)
    # Allow passing this in as AIOCOAP_DEFAULTS_EXPECT_ALL=1 via the
    # environment, as that's easier to set in tox
    p.add_argument("--expect-all", help="Exit with an error unless all subsystems are available",
            action='store_true', default=os.environ.get('AIOCOAP_DEFAULTS_EXPECT_ALL') == '1')
    p.add_argument('--version', action='version', version=version)
    args = p.parse_args()

    error = 0

    print("Python version: %s" % sys.version)
    print("aiocoap version: %s" % version)
    print("Modules missing for subsystems:")
    for (name, f) in missing_module_functions.items():
        missing = f()
        if missing and args.expect_all:
            error = 1
        print("    %s: %s" % (name, "everything there" if not missing else "missing " + ", ".join(missing)))
    print("Python platform: %s" % sys.platform)
    print("Default server transports:  %s" % ":".join(get_default_servertransports(use_env=False)))
    print("Selected server transports: %s" % ":".join(get_default_servertransports()))
    print("Default client transports:  %s" % ":".join(get_default_clienttransports(use_env=False)))
    print("Selected client transports: %s" % ":".join(get_default_clienttransports()))
    print("SO_REUSEPORT available (default, selected): %s, %s" % (has_reuse_port(use_env=False), has_reuse_port()))

    if error:
        print("Exiting unsuccessfully because --expect-all was set and not all extras are available.")
    sys.exit(error)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Common code for the tcp and the ws modules, both of which are based on
RFC8323 mechanisms, but differ in their underlying protocol implementations
(asyncio stream vs. websockets module) far enough that they only share small
portions of their code"""

import asyncio
from typing import Optional
from aiocoap import Message
from aiocoap import optiontypes, util
from aiocoap.numbers.codes import CSM, PING, PONG, RELEASE, ABORT
from aiocoap import error

class CloseConnection(Exception):
    """Raised in RFC8323 common processing to trigger a connection shutdown on
    the TCP / WebSocket side.

    The TCP / WebSocket side should send the exception's argument on to the
    token manager, close the connection, and does not need to perform further
    logging."""

class RFC8323Remote:
    """Mixin for Remotes for all the common RFC8323 processing

    Implementations still need the per-transport parts, especially a
    _send_message and an _abort_with implementation.
    """

    # CSM received from the peer. The receive hook should abort suitably when
    # receiving a non-CSM message and this is not set yet.
    _remote_settings: Optional[Message]

    # Parameter usually set statically per implementation
    _my_max_message_size = 1024 * 1024

    def __init__(self):
        self._remote_settings = None

    is_multicast = False
    is_multicast_locally = False

    # implementing interfaces.EndpointAddress

    def __repr__(self):
        return "<%s at %#x, hostinfo %s, local %s>" % (type(self).__name__, id(self), self.hostinfo, self.hostinfo_local)

    @property
    def hostinfo(self):
        # keeping _remote_hostinfo and _local_hostinfo around structurally rather than in
        # hostinfo / hostinfo_local form looks odd now, but on the long run the
        # remote should be able to tell the message what its default Uri-Host
        # value is
        return util.hostportjoin(*self._remote_hostinfo)

    @property
    def hostinfo_local(self):
        return util.hostportjoin(*self._local_hostinfo)

    @property
    def uri_base(self):
        if self._local_is_server:
            raise error.AnonymousHost("Client side of %s can not be expressed as a URI" % self._ctx._scheme)
        else:
            return self._ctx._scheme + '://' + self.hostinfo

    @property
    def uri_base_local(self):
        if self._local_is_server:
            return self._ctx._scheme + '://' + self.hostinfo_local
        else:
            raise error.AnonymousHost("Client side of %s can not be expressed as a URI" % self._ctx._scheme)

    @property
    def maximum_block_size_exp(self):
        if self._remote_settings is None:
            # This is assuming that we can do BERT, so a first Block1 would be
            # exponent 7 but still only 1k -- because by the time we send this,
            # we typically haven't seen a CSM yet, so we'd be stuck with 6
            # because 7959 says we can't increase the exponent...
            #
            # FIXME: test whether we're properly using lower block sizes if
            # server says that szx=7 is not OK.
            return 7

        max_message_size = (self._remote_settings or {}).get('max-message-size', 1152)
        has_blockwise = (self._remote_settings or {}).get('block-wise-transfer', False)
        if max_message_size > 1152 and has_blockwise:
            return 7
        return 6 # FIXME: deal with smaller max-message-size

    @property
    def maximum_payload_size(self):
        # see maximum_payload_size of interfaces comment
        slack = 100

        max_message_size = (self._remote_settings or {}).get('max-message-size', 1152)
        has_blockwise = (self._remote_settings or {}).get('block-wise-transfer', False)
        if max_message_size > 1152 and has_blockwise:
            return ((max_message_size - 128) // 1024) * 1024 + slack
        return 1024 + slack # FIXME: deal with smaller max-message-size

    @property
    def blockwise_key(self):
        return (self._remote_hostinfo, self._local_hostinfo)

    # Utility methods for implementing an RFC8323 transport

    def _send_initial_csm(self):
        my_csm = Message(code=CSM)
        # this is a tad awkward in construction because the options objects
        # were designed under the assumption that the option space is constant
        # for all message codes.
        block_length = optiontypes.UintOption(2, self._my_max_message_size)
        my_csm.opt.add_option(block_length)
        supports_block = optiontypes.UintOption(4, 0)
        my_csm.opt.add_option(supports_block)
        self._send_message(my_csm)

    def _process_signaling(self, msg):
        if msg.code == CSM:
            if self._remote_settings is None:
                self._remote_settings = {}
            for opt in msg.opt.option_list():
                # FIXME: this relies on the relevant option numbers to be
                # opaque; message parsing should already use the appropriate
                # option types, or re-think the way options are parsed
                if opt.number == 2:
                    self._remote_settings['max-message-size'] = int.from_bytes(opt.value, 'big')
                elif opt.number == 4:
                    self._remote_settings['block-wise-transfer'] = True
                elif opt.number.is_critical():
                    self.abort("Option not supported", bad_csm_option=opt.number)
                else:
                    pass # ignoring elective CSM options
        elif msg.code in (PING, PONG, RELEASE, ABORT):
            # not expecting data in any of them as long as Custody is not implemented
            for opt in msg.opt.option_list():
                if opt.number.is_critical():
                    self.abort("Unknown critical option")
                else:
                    pass

            if msg.code == PING:
                pong = Message(code=PONG, token=msg.token)
                self._send_message(pong)
            elif msg.code == PONG:
                pass
            elif msg.code == RELEASE:
                # The behavior SHOULD be enhanced to answer outstanding
                # requests, but it is unclear to which extent this side may
                # still use the connection.
                self.log.info("Received Release, closing on this end (options: %s)", msg.opt)
                raise CloseConnection(error.RemoteServerShutdown("Peer released connection"))
            elif msg.code == ABORT:
                self.log.warning("Received Abort (options: %s)", msg.opt)
                raise CloseConnection(error.RemoteServerShutdown("Peer aborted connection"))
        else:
            self.abort("Unknown signalling code")

    def abort(self, errormessage=None, bad_csm_option=None):
        self.log.warning("Aborting connection: %s", errormessage)
        abort_msg = Message(code=ABORT)
        if errormessage is not None:
            abort_msg.payload = errormessage.encode('utf8')
        if bad_csm_option is not None:
            bad_csm_option_option = optiontypes.UintOption(2, bad_csm_option)
            abort_msg.opt.add_option(bad_csm_option_option)
        self._abort_with(abort_msg)

    async def release(self):
        """Send Release message, (not implemented:) wait for connection to be
        actually closed by the peer.

        Subclasses should extend this to await closing of the connection,
        especially if they'd get into lock-up states otherwise (was would
        WebSockets).
        """
        self.log.info("Releasing connection %s", self)
        release_msg = Message(code=RELEASE)
        self._send_message(release_msg)

        try:
            # FIXME: we could wait for the peer to close the connection, but a)
            # that'll need some work on the interface between this module and
            # ws/tcp, and b) we have no peers to test this with that would
            # produce any sensible data (as aiocoap on release just closes).
            pass
        except asyncio.CancelledError:
            self.log.warning(
                    "Connection %s was not closed by peer in time after release",
                    self
                    )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module implements a MessageInterface for UDP based on the asyncio
DatagramProtocol.

This is a simple version that works only for servers bound to a single unicast
address. It provides a server backend in situations when :mod:`.udp6` is
unavailable and :mod:`.simple6` needs to be used for clients.

While it is in theory capable of sending requests too, it should not be used
like that, because it won't receive ICMP errors (see below).

Shortcomings
------------

* This implementation does not receive ICMP errors. This violates the CoAP
  standard and can lead to unnecessary network traffic, bad user experience
  (when used for client requests) or even network attack amplification.

* The server can not be used with the "any-address" (``::``, ``0.0.0.0``).
  If it were allowed to bind there, it would not receive any indication from the operating system
  as to which of its own addresses a request was sent,
  and could not send the response with the appropriate sender address.

  (The :mod:`udp6<aiocoap.transports.udp6>` transport does not suffer that shortcoming,
  simplesocketserver is typically only used when that is unavailable).

  With simplesocketserver, you need to explicitly give the IP address of your server
  in the ``bind`` argument of :meth:`aiocoap.protocol.Context.create_server_context`.

* This transport is experimental and likely to change.
"""

import asyncio
from collections import namedtuple

from .. import error
from ..numbers import COAP_PORT
from .. import interfaces
from .generic_udp import GenericMessageInterface
from ..util import hostportjoin
from .. import defaults

class _Address(namedtuple('_Address', ['serversocket', 'address']), interfaces.EndpointAddress):
    # hashability and equality follow from being a namedtuple
    def __repr__(self):
        return '<%s.%s via %s to %s>' % (__name__, type(self).__name__, self.serversocket, self.address)

    def send(self, data):
        self.serversocket._transport.sendto(data, self.address)

    # EnpointAddress interface

    is_multicast = False
    is_multicast_locally = False

    @property
    def hostinfo(self):
        # `host` already contains the interface identifier, so throwing away
        # scope and interface identifier
        host, port, *_ = self.address
        if port == COAP_PORT:
            port = None
        return hostportjoin(host, port)

    @property
    def uri_base(self):
        return self.scheme + '://' + self.hostinfo

    @property
    def hostinfo_local(self):
        return self.serversocket.hostinfo_local

    @property
    def uri_base_local(self):
        return self.scheme + '://' + self.hostinfo_local

    scheme = 'coap'

    @property
    def blockwise_key(self):
        return self.address

class _DatagramServerSocketSimple(asyncio.DatagramProtocol):
    # To be overridden by tinydtls_server
    _Address = _Address

    @classmethod
    async def create(cls, bind, log, loop, message_interface: "GenericMessageInterface"):
        if bind is None or bind[0] in ('::', '0.0.0.0', '', None):
            # If you feel tempted to remove this check, think about what
            # happens if two configured addresses can both route to a
            # requesting endpoint, how that endpoint is supposed to react to a
            # response from the other address, and if that case is not likely
            # to ever happen in your field of application, think about what you
            # tell the first user where it does happen anyway.
            raise ValueError("The transport can not be bound to any-address.")

        ready = asyncio.get_running_loop().create_future()

        transport, protocol = await loop.create_datagram_endpoint(
                lambda: cls(ready.set_result, message_interface, log),
                local_addr=bind,
                reuse_port=defaults.has_reuse_port(),
                )

        # Conveniently, we only bind to a single port (because we need to know
        # the return address, not because we insist we know the local
        # hostinfo), and can thus store the local hostinfo without distinction
        protocol.hostinfo_local = hostportjoin(bind[0], bind[1] if bind[1] != COAP_PORT else None)

        self = await ready
        self._loop = loop
        return self

    def __init__(self, ready_callback, message_interface: "GenericMessageInterface", log):
        self._ready_callback = ready_callback
        self._message_interface = message_interface
        self.log = log

    async def shutdown(self):
        self._transport.abort()

    # interface like _DatagramClientSocketpoolSimple6

    async def connect(self, sockaddr):
        # FIXME this is not regularly tested either

        self.log.warning("Sending initial messages via a server socket is not recommended")
        # A legitimate case is when something stores return addresses as
        # URI(part)s and not as remotes. (In similar transports this'd also be
        # the case if the address's connection is dropped from the pool, but
        # that doesn't happen here since there is no pooling as there is no
        # per-connection state).

        # getaddrinfo is not only to needed to resolve any host names (which
        # would not be recognized otherwise), but also to get a complete (host,
        # port, zoneinfo, whatwasthefourth) tuple from what is passed in as a
        # (host, port) tuple.
        addresses = await self._loop.getaddrinfo(*sockaddr, family=self._transport.get_extra_info('socket').family)
        if not addresses:
            raise error.NetworkError("No addresses found for %s" % sockaddr[0])
        # FIXME could do happy eyebals
        address = addresses[0][4]
        address = self._Address(self, address)
        return address

    # datagram protocol interface

    def connection_made(self, transport):
        self._transport = transport
        self._ready_callback(self)
        del self._ready_callback

    def datagram_received(self, data, sockaddr):
        self._message_interface._received_datagram(self._Address(self, sockaddr), data)

    def error_received(self, exception):
        # This is why this whole implementation is a bad idea (but still the best we got on some platforms)
        self.log.warning("Ignoring error because it can not be mapped to any connection: %s", exception)

    def connection_lost(self, exception):
        if exception is None:
            pass # regular shutdown
        else:
            self.log.error("Received unexpected connection loss: %s", exception)

class MessageInterfaceSimpleServer(GenericMessageInterface):
    # for alteration by tinydtls_server
    _default_port = COAP_PORT
    _serversocket = _DatagramServerSocketSimple

    @classmethod
    async def create_server(cls, bind, ctx: interfaces.MessageManager, log, loop):
        self = cls(ctx, log, loop)
        bind = bind or ('::', None)
        # Interpret None as 'default port', but still allow to bind to 0 for
        # servers that want a random port (eg. when the service URLs are
        # advertised out-of-band anyway). LwM2M clients should use simple6
        # instead as outlined there.
        bind = (bind[0], self._default_port if bind[1] is None else bind[1] + (self._default_port - COAP_PORT))

        # Cyclic reference broken during shutdown
        self._pool = await self._serversocket.create(bind, log, self._loop, self)

        return self

    async def recognize_remote(self, remote):
        # FIXME: This is never tested (as is the connect method) because all
        # tests create client contexts client-side (which don't build a
        # simplesocketserver), and because even when a server context is
        # created, there's a simple6 that grabs such addresses before a request
        # is sent out
        return isinstance(remote, _Address) and remote.serversocket is self._pool
